import tensorflow as tf
import torch
from tensorflow.keras.layers import Layer
from tensorflow.python.keras.losses import CategoricalCrossentropy
from tensorflow_ranking.python.keras.losses import PairwiseHingeLoss

_hinge_loss = torch.nn.MarginRankingLoss(margin=1, reduction="mean")


class KerasPairModel(tf.keras.Model):
    def __init__(self, model, *args, **kwargs):
        super(KerasPairModel, self).__init__(*args, **kwargs)
        self.model = model

    def call(self, x, **kwargs):
        score = self.model.score(x, **kwargs)
        return score

    def predict_step(self, data):
        return self.model.predict_step(data)


class KerasTripletModel(tf.keras.Model):
    def __init__(self, model, *args, **kwargs):
        super(KerasTripletModel, self).__init__(*args, **kwargs)
        self.model = model

    def call(self, x, **kwargs):
        pos_score, neg_score = self.model.score_pair(x, **kwargs)
        stacked_score = tf.stack([pos_score, neg_score], axis=1)

        return stacked_score

    def predict_step(self, data):
        return self.model.predict_step(data)


class TFPairwiseHingeLoss(PairwiseHingeLoss):
    def call(self, y_true, y_pred):
        y_true = tf.reshape(y_true, [-1, 2])
        y_pred = tf.reshape(y_pred, [-1, 2])

        return super(TFPairwiseHingeLoss, self).call(y_true, y_pred)


class TFCategoricalCrossEntropyLoss(CategoricalCrossentropy):
    def call(self, ytrue, ypred):
        tf.debugging.assert_equal(tf.shape(ytrue), tf.shape(ypred))

        return super(TFCategoricalCrossEntropyLoss, self).call(ytrue, ypred)


def pair_softmax_loss(pos_neg_scores, *args, **kwargs):
    scores = torch.stack(pos_neg_scores, dim=1)
    return torch.mean(1.0 - scores.softmax(dim=1)[:, 0])


def pair_hinge_loss(pos_neg_scores, *args, **kwargs):
    label = torch.ones_like(pos_neg_scores[0])  # , dtype=torch.int)
    return _hinge_loss(pos_neg_scores[0], pos_neg_scores[1], label)


def similarity_matrix_tf(query_embed, doc_embed, query_tok, doc_tok, padding):
    batch_size, qlen, doclen = tf.shape(query_embed)[0], tf.shape(query_embed)[1], tf.shape(doc_embed)[1]
    q_denom = tf.broadcast_to(tf.reshape(tf.norm(query_embed, axis=2), (batch_size, qlen, 1)), (batch_size, qlen, doclen)) + 1e-9
    doc_denom = (
        tf.broadcast_to(tf.reshape(tf.norm(doc_embed, axis=2), (batch_size, 1, doclen)), (batch_size, qlen, doclen)) + 1e-9
    )

    # Why perm?
    # let query have shape (32, 8, 300)
    # let doc have shape (32, 800, 300)
    # Our similarity matrix should have the shape (32, 8, 800)
    # The perm is required so that the result of matmul will have this shape
    perm = tf.transpose(doc_embed, perm=[0, 2, 1])
    sim = tf.matmul(query_embed, perm) / (q_denom * doc_denom)
    nul = tf.zeros_like(sim)
    sim = tf.where(tf.broadcast_to(tf.reshape(query_tok, (batch_size, qlen, 1)), (batch_size, qlen, doclen)) == padding, nul, sim)
    sim = tf.where(tf.broadcast_to(tf.reshape(doc_tok, (batch_size, 1, doclen)), (batch_size, qlen, doclen)) == padding, nul, sim)

    # TODO: Add support for handling list inputs (eg: for CEDR). See the pytorch implementation of simmat
    return sim


class SimilarityMatrix(torch.nn.Module):
    def __init__(self, embedding):
        super().__init__()
        self.embedding = embedding
        self.padding = 0

    def remove_padding(self, sim, query_tok, doc_tok, BAT, A, B):
        nul = torch.zeros_like(sim)
        sim = torch.where(query_tok.reshape(BAT, A, 1).expand(BAT, A, B) == self.padding, nul, sim)
        sim = torch.where(doc_tok.reshape(BAT, 1, B).expand(BAT, A, B) == self.padding, nul, sim)
        return sim

    def exact_match_matrix(self, query_tok, doc_tok, BAT, A, B):
        sim = (query_tok.reshape(BAT, A, 1).expand(BAT, A, B) == doc_tok.reshape(BAT, 1, B).expand(BAT, A, B)).float()
        sim = self.remove_padding(sim, query_tok, doc_tok, BAT, A, B)
        return sim

    def cosine_similarity_matrix(self, query_tok, doc_tok, BAT, A, B):
        a_emb, b_emb = self.embedding(query_tok), self.embedding(doc_tok)
        a_denom = a_emb.norm(p=2, dim=2).reshape(BAT, A, 1).expand(BAT, A, B) + 1e-9  # avoid 0div
        b_denom = b_emb.norm(p=2, dim=2).reshape(BAT, 1, B).expand(BAT, A, B) + 1e-9  # avoid 0div
        perm = b_emb.permute(0, 2, 1)
        sim = a_emb.bmm(perm) / (a_denom * b_denom)
        sim = self.remove_padding(sim, query_tok, doc_tok, BAT, A, B)
        return sim

    # query_tok and doc_tok should contain integers
    def forward(self, query_tok, doc_tok):
        BAT, A, B = query_tok.shape[0], query_tok.shape[1], doc_tok.shape[1]
        assert doc_tok.shape[0] == BAT

        # note: all OOV terms are given negative indices and 0 is padding
        # approach:
        # 1. we calculate an exact match matrix on OOV terms only, and set padding to 0
        # 2. we calculate a cosine sim matrix on in-vocab terms only, and set padding to 0
        # 3. we sum the two matrices
        exact_match = self.exact_match_matrix(query_tok.clamp(max=0), doc_tok.clamp(max=0), BAT, A, B)
        cos_matrix = self.cosine_similarity_matrix(query_tok.clamp(min=0), doc_tok.clamp(min=0), BAT, A, B)
        simmat = exact_match + cos_matrix
        return simmat


class StackedSimilarityMatrix(torch.nn.Module):
    # based on SimmatModule from https://github.com/Georgetown-IR-Lab/cedr/blob/master/modeling_util.py
    # which is copyright (c) 2019 Georgetown Information Retrieval Lab, MIT license
    def __init__(self, padding=0):
        super().__init__()
        self.padding = padding

    # query_embed and doc_embed can be a list (eg for CEDR)
    def forward(self, query_embed, doc_embed, query_tok, doc_tok):
        simmat = []

        assert type(query_embed) == type(doc_embed)
        if not isinstance(query_embed, list):
            query_embed, doc_embed = [query_embed], [doc_embed]

        for a_emb, b_emb in zip(query_embed, doc_embed):
            BAT, A, B = a_emb.shape[0], a_emb.shape[1], b_emb.shape[1]

            if a_emb is None and b_emb is None:
                # exact match matrix
                sim = query_tok.reshape(BAT, A, 1).expand(BAT, A, B) == doc_tok.reshape(BAT, 1, B).expand(BAT, A, B).float()
            else:
                # cosine similarity matrix
                a_denom = a_emb.norm(p=2, dim=2).reshape(BAT, A, 1).expand(BAT, A, B) + 1e-9  # avoid 0div
                b_denom = b_emb.norm(p=2, dim=2).reshape(BAT, 1, B).expand(BAT, A, B) + 1e-9  # avoid 0div
                perm = b_emb.permute(0, 2, 1)
                sim = a_emb.bmm(perm) / (a_denom * b_denom)

            # set similarity values to 0 for <pad> tokens in query and doc (indicated by self.padding)
            nul = torch.zeros_like(sim)
            sim = torch.where(query_tok.reshape(BAT, A, 1).expand(BAT, A, B) == self.padding, nul, sim)
            sim = torch.where(doc_tok.reshape(BAT, 1, B).expand(BAT, A, B) == self.padding, nul, sim)

            simmat.append(sim)
        return torch.stack(simmat, dim=1)


class RbfKernel(torch.nn.Module):
    # based on KNRMRbfKernel from https://github.com/Georgetown-IR-Lab/cedr/blob/master/modeling_util.py
    # which is copyright (c) 2019 Georgetown Information Retrieval Lab, MIT license
    def __init__(self, initial_mu, initial_sigma, requires_grad=True):
        super().__init__()
        self.mu = torch.nn.Parameter(torch.tensor(initial_mu), requires_grad=requires_grad)
        self.sigma = torch.nn.Parameter(torch.tensor(initial_sigma), requires_grad=requires_grad)

    def forward(self, data):
        adj = data - self.mu
        return torch.exp(-0.5 * adj * adj / self.sigma / self.sigma)


class RbfKernelBank(torch.nn.Module):
    # based on KNRMRbfKernelBank from https://github.com/Georgetown-IR-Lab/cedr/blob/master/modeling_util.py
    # which is copyright (c) 2019 Georgetown Information Retrieval Lab, MIT license
    def __init__(self, mus=None, sigmas=None, dim=1, requires_grad=True):
        super().__init__()
        self.dim = dim
        kernels = [RbfKernel(m, s, requires_grad=requires_grad) for m, s in zip(mus, sigmas)]
        self.kernels = torch.nn.ModuleList(kernels)

    def count(self):
        return len(self.kernels)

    def forward(self, data):
        return torch.stack([k(data) for k in self.kernels], dim=self.dim)


class RbfKernelBankTF(Layer):
    def __init__(self, mus, sigmas, dim=1, requires_grad=True, **kwargs):
        super(RbfKernelBankTF, self).__init__(**kwargs)
        self.dim = dim
        self.kernel_list = [RbfKernelTF(m, s, requires_grad=requires_grad) for m, s in zip(mus, sigmas)]

    def count(self):
        return len(self.kernel_list)

    def call(self, data, **kwargs):
        return tf.stack([self.kernel_list[i](data) for i in range(len(self.kernel_list))], axis=self.dim)


class RbfKernelTF(Layer):
    def __init__(self, initial_mu, initial_sigma, requires_grad=True, **kwargs):
        super(RbfKernelTF, self).__init__(**kwargs)
        self.mu = tf.Variable(initial_mu, trainable=requires_grad, name="mus", dtype=tf.float32)
        self.sigma = tf.Variable(initial_sigma, trainable=requires_grad, name="sigmas", dtype=tf.float32)

    def call(self, data, *kwargs):
        adj = data - self.mu
        return tf.exp(-0.5 * adj * adj / self.sigma / self.sigma)


def create_emb_layer(weights, non_trainable=True):
    layer = torch.nn.Embedding(*weights.shape)
    layer.load_state_dict({"weight": torch.tensor(weights)})

    if non_trainable:
        layer.weight.requires_grad = False
    else:
        layer.weight.requires_grad = True

    return layer



import pickle
from zipfile import ZipFile

from tqdm import tqdm

from capreolus import ConfigOption, constants
from capreolus.utils.common import download_file, remove_newline
from capreolus.utils.loginit import get_logger
from capreolus.utils.trec import document_to_trectxt

from . import Collection

logger = get_logger(__name__)
PACKAGE_PATH = constants["PACKAGE_PATH"]


@Collection.register
class CodeSearchNet(Collection):
    """CodeSearchNet Corpus. [1]

    [1] Hamel Husain, Ho-Hsiang Wu, Tiferet Gazit, Miltiadis Allamanis, and Marc Brockschmidt. 2019. CodeSearchNet Challenge: Evaluating the State of Semantic Code Search. arXiv 2019.
    """

    module_name = "codesearchnet"
    url = "https://s3.amazonaws.com/code-search-net/CodeSearchNet/v2"
    collection_type = "TrecCollection"  # TODO: any other supported type?
    generator_type = "DefaultLuceneDocumentGenerator"
    config_spec = [ConfigOption("lang", "ruby", "CSN language dataset to use")]

    def download_if_missing(self):
        cachedir = self.get_cache_path()
        document_dir = cachedir / "documents"
        coll_filename = document_dir / ("csn-" + self.config["lang"] + "-collection.txt")

        if coll_filename.exists():
            return document_dir.as_posix()

        zipfile = self.config["lang"] + ".zip"
        lang_url = f"{self.url}/{zipfile}"
        tmp_dir = cachedir / "tmp"
        zip_path = tmp_dir / zipfile

        if zip_path.exists():
            logger.info(f"{zipfile} already exist under directory {tmp_dir}, skip downloaded")
        else:
            tmp_dir.mkdir(exist_ok=True, parents=True)
            download_file(lang_url, zip_path)

        document_dir.mkdir(exist_ok=True, parents=True)  # tmp
        with ZipFile(zip_path, "r") as zipobj:
            zipobj.extractall(tmp_dir)

        pkl_path = tmp_dir / (self.config["lang"] + "_dedupe_definitions_v2.pkl")
        self._pkl2trec(pkl_path, coll_filename)
        return document_dir.as_posix()

    def _pkl2trec(self, pkl_path, trec_path):
        lang = self.config["lang"]
        with open(pkl_path, "rb") as f:
            codes = pickle.load(f)

        fout = open(trec_path, "w", encoding="utf-8")
        for i, code in tqdm(enumerate(codes), desc=f"Preparing the {lang} collection file"):
            docno = f"{lang}-FUNCTION-{i}"
            doc = remove_newline(" ".join(code["function_tokens"]))
            fout.write(document_to_trectxt(docno, doc))
        fout.close()



import tensorflow as tf
from transformers import TFAutoModelForSequenceClassification

from capreolus import ConfigOption, Dependency
from capreolus.reranker import Reranker


class TFElectraRelevanceHead(tf.keras.layers.Layer):
    """ BERT-style ClassificationHead (i.e., out_proj only -- no dense). See transformers.TFElectraClassificationHead """

    def __init__(self, dropout, out_proj, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dropout = dropout
        self.out_proj = out_proj

    def call(self, inputs, **kwargs):
        x = inputs[:, 0, :]  # take <s> token (equiv. to [CLS])
        x = self.dropout(x)
        x = self.out_proj(x)
        return x


class TFBERTMaxP_Class(tf.keras.layers.Layer):
    def __init__(self, extractor, config, *args, **kwargs):
        super(TFBERTMaxP_Class, self).__init__(*args, **kwargs)
        self.extractor = extractor

        if config["pretrained"] == "electra-base-msmarco":
            self.bert = TFAutoModelForSequenceClassification.from_pretrained("Capreolus/electra-base-msmarco")
            dropout, fc = self.bert.classifier.dropout, self.bert.classifier.out_proj
            self.bert.classifier = TFElectraRelevanceHead(dropout, fc)
        elif config["pretrained"] == "bert-base-msmarco":
            self.bert = TFAutoModelForSequenceClassification.from_pretrained("Capreolus/bert-base-msmarco")
        elif config["pretrained"] == "bert-base-uncased":
            self.bert = TFAutoModelForSequenceClassification.from_pretrained(config["pretrained"], hidden_dropout_prob=0.1)
        else:
            raise ValueError(
                f"unsupported model: {config['pretrained']}; need to ensure correct tokenizers will be used before arbitrary hgf models are supported"
            )

        self.config = config

    def call(self, x, **kwargs):
        """
        Returns logits of shape [2]
        """
        doc_bert_input, doc_mask, doc_seg = x[0], x[1], x[2]

        passage_scores = self.bert(doc_bert_input, attention_mask=doc_mask, token_type_ids=doc_seg)[0]

        return passage_scores

    def predict_step(self, data):
        """
        Scores each passage and applies max pooling over it.
        """
        posdoc_bert_input, posdoc_mask, posdoc_seg, negdoc_bert_input, negdoc_mask, negdoc_seg = data
        batch_size = tf.shape(posdoc_bert_input)[0]
        num_passages = self.extractor.config["numpassages"]
        maxseqlen = self.extractor.config["maxseqlen"]

        posdoc_bert_input = tf.reshape(posdoc_bert_input, [batch_size * num_passages, maxseqlen])
        posdoc_mask = tf.reshape(posdoc_mask, [batch_size * num_passages, maxseqlen])
        posdoc_seg = tf.reshape(posdoc_seg, [batch_size * num_passages, maxseqlen])

        passage_scores = self.call((posdoc_bert_input, posdoc_mask, posdoc_seg), training=False)[:, 1]
        passage_scores = tf.reshape(passage_scores, [batch_size, num_passages])

        if self.config["aggregation"] == "max":
            passage_scores = tf.math.reduce_max(passage_scores, axis=1)
        elif self.config["aggregation"] == "first":
            passage_scores = passage_scores[:, 0]
        elif self.config["aggregation"] == "sum":
            passage_scores = tf.math.reduce_sum(tf.nn.softmax(passage_scores), axis=1)
        else:
            raise ValueError("Unknown aggregation method: {}".format(self.config["aggregation"]))

        return passage_scores

    def score(self, x, **kwargs):
        posdoc_bert_input, posdoc_mask, posdoc_seg, negdoc_bert_input, negdoc_mask, negdoc_seg = x

        return self.call((posdoc_bert_input, posdoc_mask, posdoc_seg), **kwargs)

    def score_pair(self, x, **kwargs):
        posdoc_bert_input, posdoc_mask, posdoc_seg, negdoc_bert_input, negdoc_mask, negdoc_seg = x

        pos_score = self.call((posdoc_bert_input, posdoc_mask, posdoc_seg), **kwargs)[:, 1]
        neg_score = self.call((negdoc_bert_input, negdoc_mask, negdoc_seg), **kwargs)[:, 1]

        return pos_score, neg_score


@Reranker.register
class TFBERTMaxP(Reranker):
    module_name = "TFBERTMaxP"

    dependencies = [
        Dependency(key="extractor", module="extractor", name="bertpassage"),
        Dependency(key="trainer", module="trainer", name="tensorflow"),
    ]
    config_spec = [
        ConfigOption(
            "pretrained", "bert-base-uncased", "Pretrained model: bert-base-uncased, bert-base-msmarco, or electra-base-msmarco"
        ),
        ConfigOption("aggregation", "max"),
    ]

    def build_model(self):
        self.model = TFBERTMaxP_Class(self.extractor, self.config)
        return self.model



from capreolus import ConfigOption, Dependency, evaluator
from capreolus.task import Task
from capreolus.utils.loginit import get_logger

logger = get_logger(__name__)  # pylint: disable=invalid-name


@Task.register
class TutorialTask(Task):
    module_name = "tutorial"
    config_spec = [ConfigOption("optimize", "map", "metric to maximize on the validation set")]
    dependencies = [
        Dependency(key="benchmark", module="benchmark", name="nf", provide_this=True, provide_children=["collection"]),
        Dependency(key="searcher1", module="searcher", name="BM25RM3"),
        Dependency(key="searcher2", module="searcher", name="SDM"),
    ]

    commands = ["run"] + Task.help_commands
    default_command = "run"

    def run(self):
        output_dir = self.get_results_path()

        # read the title queries from the chosen benchmark's topic file
        results1 = self.searcher1.query_from_file(self.benchmark.topic_file, output_dir / "searcher1")
        results2 = self.searcher2.query_from_file(self.benchmark.topic_file, output_dir / "searcher2")
        searcher_results = [results1, results2]

        # using the benchmark's folds, which each contain train/validation/test queries,
        # choose the best run in `output_dir` for the fold based on the validation queries
        # and return metrics calculated on the test queries
        best_results = evaluator.search_best_run(
            searcher_results, self.benchmark, primary_metric=self.config["optimize"], metrics=evaluator.DEFAULT_METRICS
        )

        for fold, path in best_results["path"].items():
            shortpath = "..." + path[-40:]
            logger.info("fold=%s best run: %s", fold, shortpath)

        logger.info("cross-validated results when optimizing for '%s':", self.config["optimize"])
        for metric, score in sorted(best_results["score"].items()):
            logger.info("%15s: %0.4f", metric, score)

        return best_results



import contextlib
import numpy as np
import torch
from torch import nn
from transformers import BertForNextSentencePrediction

from capreolus import ConfigOption, Dependency
from capreolus.reranker import Reranker
from capreolus.utils.loginit import get_logger

logger = get_logger(__name__)


# official weights converted with:
# def convert(name):
#     from transformers import BertTokenizer, BertForNextSentencePrediction, TFBertForNextSentencePrediction

#     tokenizer = BertTokenizer.from_pretrained("bert-large-uncased")

#     state = torch.load(f"/GW/NeuralIR/nobackup/birch-emnlp_bert4ir_v2/models/saved.{name}_1", map_location="cpu")

#     model = BertForNextSentencePrediction.from_pretrained("bert-large-uncased")
#     model.load_state_dict(state["model"].state_dict())

#     output = f"/GW/NeuralIR/nobackup/birch-emnlp_bert4ir_v2/models/export/birch-bert-large-{name}"
#     os.makedirs(output, exist_ok=True)
#     model.save_pretrained(output)
#     tokenizer.save_pretrained(output)

#     # tf2 support
#     tf_model = TFBertForNextSentencePrediction.from_pretrained(output, from_pt=True)
#     tf_model.save_pretrained(output)


class Birch_Class(nn.Module):
    def __init__(self, extractor, config):
        super().__init__()

        self.config = config

        if config["hidden"] == 0:
            self.combine = nn.Linear(config["topk"], 1, bias=False)
            with torch.no_grad():
                self.combine.weight = nn.Parameter(torch.ones_like(self.combine.weight) / config["topk"])
        else:
            assert config["hidden"] > 0
            self.combine = nn.Sequential(nn.Linear(config["topk"], config["hidden"]), nn.ReLU(), nn.Linear(config["hidden"], 1))

        # original model file (requires apex):
        # state = torch.load("/GW/NeuralIR/nobackup/birch-emnlp_bert4ir_v2/models/saved.msmarco_mb_1", map_location="cpu")
        # self.bert = state["model"]

        # saved.msmarco_mb_1 weights exported from the official apex model:
        # self.bert = BertForNextSentencePrediction.from_pretrained("bert-large-uncased")
        # self.bert.load_state_dict(torch.load("/GW/NeuralIR/nobackup/birch-emnlp_bert4ir_v2/models/converted"))
        # converted_weights.msmarco_mb

        # kevin's base model:
        # self.bert = BertForNextSentencePrediction.from_pretrained("bert-base-uncased")
        # saved_bert = torch.load("/GW/NeuralIR/nobackup/birch/models/saved.tmp_1")["model"]
        # self.bert.load_state_dict(saved_bert.state_dict())

        # also /GW/NeuralIR/nobackup/birch-emnlp_bert4ir_v2/models/export/birch-bert-base-kevin
        self.bert = BertForNextSentencePrediction.from_pretrained(f"Capreolus/birch-bert-large-{config['pretrained']}")

        if not config["finetune"]:
            self.bert.requires_grad = False
            self.bert_context = torch.no_grad
        else:
            self.bert_context = contextlib.nullcontext

    def forward(self, doc, seg, mask):
        batch = doc.shape[0]

        with self.bert_context():
            bi_scores = [self.score_passages(doc[bi], seg[bi], mask[bi], batch) for bi in range(batch)]
            scores = torch.stack(bi_scores)
            assert scores.shape == (batch, self.config["extractor"]["numpassages"], 2)
            scores = scores[:, :, 1]  # take second output

        topk, _ = torch.topk(scores, dim=1, k=self.config["topk"])
        doc_score = self.combine(topk)
        return doc_score

    def score_passages(self, doc, seg, mask, batch):
        needed_passages = doc.shape[0]
        maxlen = doc.shape[-1]

        # find instances that contain a document (segment B)
        # for unmasked tokens in seg B, seg+mask=2
        # there are always two SEPs in segment B, so the document is not empty if there are >= 3 tokens where seg+mask=2
        valid = ((seg + mask) == 2).sum(dim=1) > 2
        if not any(valid):
            valid[0] = True
        doc, seg, mask = doc[valid], seg[valid], mask[valid]

        out = []
        batches = np.ceil(doc.shape[0] / batch).astype(int)
        for bi in range(batches):
            start = bi * batch
            stop = (bi + 1) * batch

            sb_doc, sb_seg, sb_mask = doc[start:stop], seg[start:stop], mask[start:stop]

            # find first non-padding token and shorten batch to this length
            idx = (sb_seg + sb_mask).argmax(dim=1).max()
            sb_doc = sb_doc[:, : idx + 1]
            sb_seg = sb_seg[:, : idx + 1]
            sb_mask = sb_mask[:, : idx + 1]
            # for idx in reversed(range(maxlen)):
            #     if any(sb_mask[:, idx]):
            #         sb_doc = sb_doc[:, : idx + 1]
            #         sb_seg = sb_seg[:, : idx + 1]
            #         sb_mask = sb_mask[:, : idx + 1]
            #         break

            sb_scores = self.bert(input_ids=sb_doc, token_type_ids=sb_seg, attention_mask=sb_mask)
            sb_scores = sb_scores[0]  # for new bert output
            out.append(sb_scores)

        real_out = torch.cat(out, dim=0)
        found_passages = real_out.shape[0]
        if found_passages < needed_passages:
            pad_out = torch.min(real_out, dim=0)[0].repeat(needed_passages - found_passages, 1)
            return torch.cat((real_out, pad_out), dim=0)
        else:
            return real_out


@Reranker.register
class Birch(Reranker):
    module_name = "birch"

    config_spec = [
        ConfigOption("topk", 3, "top k scores to use"),
        ConfigOption("hidden", 0, "size of hidden layer or 0 to take the weighted sum of the topk"),
        ConfigOption("finetune", False, "fine-tune the BERT model"),
        ConfigOption("pretrained", "msmarco_mb", "pretrained Birch model to load: mb, msmarco_mb, or car_mb"),
    ]
    dependencies = [
        Dependency(
            key="extractor",
            module="extractor",
            name="bertpassage",
            default_config_overrides={"tokenizer": {"pretrained": "bert-large-uncased"}},
        ),
        Dependency(key="trainer", module="trainer", name="pytorch"),
    ]

    def build_model(self):
        self.model = Birch_Class(self.extractor, self.config)
        return self.model

    def score(self, d):
        return [
            self.model(d["pos_bert_input"], d["pos_seg"], d["pos_mask"]).view(-1),
            self.model(d["neg_bert_input"], d["neg_seg"], d["neg_mask"]).view(-1),
        ]

    def test(self, d):
        return self.model(d["pos_bert_input"], d["pos_seg"], d["pos_mask"]).view(-1)



import os

import numpy as np

from capreolus import ConfigOption, Dependency, evaluator
from capreolus.sampler import PredSampler
from capreolus.searcher import Searcher
from capreolus.task import Task
from capreolus.utils.loginit import get_logger

logger = get_logger(__name__)


@Task.register
class ReRerankTask(Task):
    module_name = "rererank"
    config_spec = [
        ConfigOption("fold", "s1", "fold to run"),
        ConfigOption("optimize", "map", "metric to maximize on the dev set"),
        ConfigOption("topn", 100, "number of stage two results to rerank"),
    ]
    dependencies = [
        Dependency(
            key="benchmark", module="benchmark", name="robust04.yang19", provide_this=True, provide_children=["collection"]
        ),
        Dependency(key="rank", module="task", name="rank", provide_this=True),
        Dependency(key="rerank1", module="task", name="rerank"),
        Dependency(key="rerank2", module="task", name="rerank"),
    ]

    commands = ["train", "evaluate", "traineval"] + Task.help_commands
    default_command = "describe"

    def traineval(self):
        self.train()
        self.evaluate()

    def train(self):
        fold = self.config["fold"]
        logger.debug("results path: %s", self.get_results_path())

        self.rank.search()
        rank_results = self.rank.evaluate()
        best_search_run_path = rank_results["path"][fold]
        best_search_run = Searcher.load_trec_run(best_search_run_path)

        second_stage_results = self.rerank1.rerank_run(best_search_run, self.rerank1.get_results_path(), include_train=True)
        second_stage_topn = {
            qid: dict(sorted(docids.items(), key=lambda x: x[1], reverse=True)[: self.config["topn"]])
            for split in ("train", "dev", "test")
            for qid, docids in second_stage_results[split].items()
        }

        third_stage_results = self.rerank2.rerank_run(second_stage_topn, self.get_results_path())
        return third_stage_results

    def evaluate(self):
        fold = self.config["fold"]
        train_output_path = self.get_results_path()
        test_output_path = train_output_path / "pred" / "test" / "best"
        logger.debug("results path: %s", train_output_path)

        if os.path.exists(test_output_path):
            test_preds = Searcher.load_trec_run(test_output_path)
        else:
            self.rank.search()
            rank_results = self.rank.evaluate()
            best_search_run_path = rank_results["path"][fold]
            best_search_run = Searcher.load_trec_run(best_search_run_path)

            docids = set(docid for querydocs in best_search_run.values() for docid in querydocs)
            self.reranker.extractor.preprocess(
                qids=best_search_run.keys(), docids=docids, topics=self.benchmark.topics[self.benchmark.query_type]
            )
            self.reranker.build_model()
            self.reranker.searcher_scores = best_search_run

            self.reranker.trainer.load_best_model(self.reranker, train_output_path)

            test_run = {
                qid: docs for qid, docs in best_search_run.items() if qid in self.benchmark.folds[fold]["predict"]["test"]
            }
            test_dataset = PredSampler()
            test_dataset.prepare(test_run, self.benchmark.qrels, self.reranker.extractor)

            test_preds = self.reranker.trainer.predict(self.reranker, test_dataset, test_output_path)

        metrics = evaluator.eval_runs(test_preds, self.benchmark.qrels, evaluator.DEFAULT_METRICS, self.benchmark.relevance_level)
        logger.info("rerank: fold=%s test metrics: %s", fold, metrics)

        print("\ncomputing metrics across all folds")
        avg = {}
        found = 0
        for fold in self.benchmark.folds:
            # TODO fix by using multiple Tasks
            from pathlib import Path

            pred_path = Path(test_output_path.as_posix().replace("fold-" + self.config["fold"], "fold-" + fold))
            if not os.path.exists(pred_path):
                print("\tfold=%s results are missing and will not be included" % fold)
                continue

            found += 1
            preds = Searcher.load_trec_run(pred_path)
            metrics = evaluator.eval_runs(preds, self.benchmark.qrels, evaluator.DEFAULT_METRICS, self.benchmark.relevance_level)
            for metric, val in metrics.items():
                avg.setdefault(metric, []).append(val)

        avg = {k: np.mean(v) for k, v in avg.items()}
        logger.info("rerank: average cross-validated metrics when choosing iteration based on '%s':", self.config["optimize"])
        for metric, score in sorted(avg.items()):
            logger.info("%25s: %0.4f", metric, score)



import os
import shutil

import pytest

from capreolus import Collection, constants, module_registry
from capreolus.tests.common_fixtures import tmpdir_as_cache

collections = set(module_registry.get_module_names("collection"))


@pytest.mark.parametrize("collection_name", collections)
def test_collection_creatable(tmpdir_as_cache, collection_name):
    collection = Collection.create(collection_name)


@pytest.mark.parametrize("collection_name", collections)
@pytest.mark.download
def test_collection_downloadable(tmpdir_as_cache, collection_name):
    collection = Collection.create(collection_name)
    path = collection.find_document_path()

    # check for /tmp to reduce the impact of an invalid constants["CACHE_BASE_PATH"]
    if path.startswith("/tmp") and path.startswith(constants["CACHE_BASE_PATH"].as_posix()):
        if os.path.exists(path):
            shutil.rmtree(path)



import torch
import torch.nn as nn

from capreolus import ConfigOption, Dependency
from capreolus.reranker import Reranker
from capreolus.reranker.common import create_emb_layer
from capreolus.utils.loginit import get_logger

logger = get_logger(__name__)  # pylint: disable=invalid-name


class LocalModel(nn.Module):
    def __init__(self, p):
        super(LocalModel, self).__init__()
        self.p = p
        if p["activation"] == "tanh":
            self.activation = nn.Tanh()
        elif p["activation"] == "relu":
            self.activation = nn.ReLU()
        else:
            raise ValueError("Unexpected activation: should be either tanh or relu")

        q_len, doc_len = p["extractor"]["maxqlen"], p["extractor"]["maxdoclen"]
        dropoutrate = p["dropoutrate"]
        self.conv = nn.Sequential(nn.Conv2d(1, p["nfilters"], (1, doc_len)), self.activation)  # (B, 1, Q, D) -> (B, H, Q, 1)

        self.ffw = nn.Sequential(
            nn.Linear(q_len * p["nfilters"], p["lmhidden"]), self.activation, nn.Dropout(dropoutrate), nn.Linear(p["lmhidden"], 1)
        )

    def exact_match(self, m1, m2):
        """
        m1: (B, len1)
        m2: (B, len2)
        """
        len1, len2 = m1.size(1), m2.size(1)
        m1_expand = torch.stack([m1] * len2, dim=2)
        m2_expand = torch.stack([m2] * len1, dim=1)
        return (m1_expand == m2_expand).float()

    def forward(self, documents, queries, query_idf):
        """
        queries: (B, nq)
        documents: (B, nd)
        query_idf: (B, nq)
        """
        lm_matrix = self.exact_match(queries, documents)  # (B, nq, nd)
        if self.p["idfweight"]:
            lm_matrix = lm_matrix * query_idf[:, :, None]

        lm_x = self.conv(lm_matrix.unsqueeze(1)).squeeze()  # (B, H1, nq)
        lm_score = self.ffw(lm_x.view(lm_x.size(0), -1))
        return lm_score


class DistributedModel(nn.Module):
    def __init__(self, extractor, p):
        super(DistributedModel, self).__init__()
        if p["activation"] == "tanh":
            self.activation = nn.Tanh()
        elif p["activation"] == "relu":
            self.activation = nn.ReLU()
        else:
            raise ValueError("Unexpected activation: should be either tanh or relu")

        self.embedding = create_emb_layer(extractor.embeddings, non_trainable=True)
        embsize = self.embedding.weight.shape[-1]
        dropoutrate = p["dropoutrate"]

        self.q_conv = nn.Sequential(
            nn.Conv2d(1, p["nfilters"], (3, embsize)),
            self.activation,
            nn.Dropout(dropoutrate),
            nn.MaxPool2d((2, 1), stride=(1, 1)),
        )

        self.q_ffw = nn.Sequential(nn.Linear(p["nfilters"], p["nfilters"]))

        # (B, 1, Q, V) -> (B, H, Q', 1)
        self.d_conv1 = nn.Sequential(
            nn.Conv2d(1, p["nfilters"], (3, embsize)),
            self.activation,
            nn.Dropout(dropoutrate),
            nn.MaxPool2d((100, 1), stride=(1, 1)),
        )

        self.d_conv2 = nn.Sequential(
            nn.Conv2d(1, p["nfilters"], (p["nfilters"], 1)),  # (B, 1, H, Q') -> (B, H, 1, Q')
            self.activation,
            nn.Dropout(dropoutrate),
        )

        self.ffw_1 = nn.Sequential(nn.Linear(p["nhidden"], 1), self.activation)
        self.ffw_2 = nn.Sequential(nn.Dropout(dropoutrate), nn.Linear(p["nfilters"], 1))

    def forward(self, documents, queries):
        # dm query
        dm_q = self.embedding(queries).unsqueeze(1)  # (B, 1, nq, D)
        dm_q = self.q_conv(dm_q).squeeze()  # (B, H)
        dm_q = self.q_ffw(dm_q)  # (B, H)

        # dm document
        dm_d = self.embedding(documents).unsqueeze(1)  # (B, 1, nd, D)
        dm_d = self.d_conv1(dm_d).squeeze()  # (B, H, 699)
        dm_d = self.d_conv2(dm_d.unsqueeze(1)).squeeze()  # (B, H, 699) -> (B, 1, H, 699) -> (B, H, 1, 699) -> (B, H, 699)

        # aggregate dm_q & dm_d
        dm_x = dm_q.unsqueeze(2) * dm_d  # (B, H, 1) * (B, H, 699)
        dm_x = self.ffw_1(dm_x).squeeze()  # -> (B, H, 1) -> (B, H)
        dm_score = self.ffw_2(dm_x)  # (B, H) -> (B, H) -> (B, 1)

        return dm_score


class DUET_class(nn.Module):
    def __init__(self, extractor, p):
        super(DUET_class, self).__init__()
        self.lm = LocalModel(p)
        self.dm = DistributedModel(extractor, p)

    def forward(self, documents, queries, query_idf):
        """
        queries: (B, nq)
        documents: (B, nd)
        """
        lm_score = self.lm(documents, queries, query_idf)
        dm_score = self.dm(documents, queries)

        return lm_score + dm_score


@Reranker.register
class DUET(Reranker):
    """Bhaskar Mitra, Fernando Diaz, and Nick Craswell. 2017. Learning to Match using Local and Distributed Representations of Text for Web Search. In WWW'17."""

    module_name = "DUET"

    dependencies = [
        Dependency(key="extractor", module="extractor", name="slowembedtext"),
        Dependency(key="trainer", module="trainer", name="pytorch"),
    ]
    config_spec = [
        ConfigOption("nfilters", 10, "number of filters for both local and distrbuted model"),
        ConfigOption("lmhidden", 30, "ffw hidden layer dimension for local model"),
        ConfigOption("nhidden", 699, "ffw hidden layer dimension for local model"),
        ConfigOption("idfweight", True, "whether to weight each query word with its idf value in local model"),
        ConfigOption("dropoutrate", 0.5, "dropout probability"),
        ConfigOption("activation", "relu", "ffw layer activation: tanh or relu"),
    ]

    def build_model(self):
        if not hasattr(self, "model"):
            self.model = DUET_class(self.extractor, self.config)
        return self.model

    def score(self, d):
        query_idf = d["query_idf"]
        query_sentence = d["query"]
        pos_sentence, neg_sentence = d["posdoc"], d["negdoc"]
        return [
            self.model(pos_sentence, query_sentence, query_idf).view(-1),
            self.model(neg_sentence, query_sentence, query_idf).view(-1),
        ]

    def test(self, d):
        query_idf = d["query_idf"]
        query_sentence = d["query"]
        pos_sentence = d["posdoc"]
        return self.model(pos_sentence, query_sentence, query_idf).view(-1)



import torch
import torch.nn as nn
import torch.nn.functional as F

from capreolus import ConfigOption
from capreolus.reranker import Reranker
from capreolus.reranker.common import SimilarityMatrix, create_emb_layer
from capreolus.utils.loginit import get_logger

logger = get_logger(__name__)  # pylint: disable=invalid-name


class DRMM_class(nn.Module):
    def __init__(self, extractor, config):
        super(DRMM_class, self).__init__()
        # self.p = config
        self.nbins = config["nbins"]
        self.nodes = config["nodes"]
        self.hist_type = config["histType"]
        self.gate_type = config["gateType"]

        self.embedding = create_emb_layer(extractor.embeddings, non_trainable=True)
        self.simmat = SimilarityMatrix(self.embedding)

        self.ffw = nn.Sequential(nn.Linear(self.nbins + 1, self.nodes), nn.Tanh(), nn.Linear(self.nodes, 1), nn.Tanh())

        emb_dim = self.embedding.weight.size(-1)
        if self.gate_type == "IDF":
            self.gates = nn.Linear(1, 1, bias=False)  # for idf scalar
        elif self.gate_type == "TV":
            self.gates = nn.Linear(emb_dim, 1, bias=False)
        else:
            raise ValueError("Invalid value for gateType: gateType should be either IDF or TV")
        self.output_layer = nn.Linear(1, 1)

        # initialize FC and gate weight in the same way as MatchZoo
        nn.init.uniform_(self.ffw[0].weight, -0.1, 0.1)
        nn.init.uniform_(self.ffw[2].weight, -0.1, 0.1)
        nn.init.uniform_(self.gates.weight, -0.01, 0.01)

    def _hist_map(self, queries, documents, d_masks):
        """
        Args:
            queries: (B, Tq, H)
            documents: (B, Td, H)
            d_masks: (B, Td)

        Returns: (B, Tq, 6)
        """

        # compute cos similarity
        # q_norm = torch.sqrt((queries * queries).sum(dim=-1) + 1e-7)[:, :, None] + 1e-7
        # d_norm = torch.sqrt((documents * documents).sum(dim=-1) + 1e-7)[:, None, :] + 1e-7
        # sim_matrix = torch.bmm(queries, documents.transpose(2, 1))  # (B, Tq, Td)
        # sim_matrix = sim_matrix / q_norm
        # sim_matrix = sim_matrix / d_norm  # (B, Tq, Td)

        sim_matrix = self.simmat(queries, documents)
        sim_matrix += (1 - d_masks[:, None, :]) * 1e7  # assign large number on <PAD> pos
        hist = torch.zeros([sim_matrix.size(0), sim_matrix.size(1), self.nbins + 1], dtype=torch.float)

        idxs = list(range(self.nbins))
        bin_upperbounds = torch.linspace(-1, 1, self.nbins + 1)[1:].to(queries.device)
        for i, bin_upperbound in zip(idxs, bin_upperbounds):
            hist[:, :, i] = (sim_matrix < bin_upperbound).sum(dim=-1)
        hist[:, :, -1] = ((sim_matrix > 0.999) * (sim_matrix < 1.001)).sum(dim=-1)

        for i in list(range(self.nbins - 1, 0, -1)):  # exclude idx=config['nbins'] and idx=0
            hist[:, :, i] -= hist[:, :, i - 1]

        hist += 1

        if self.hist_type == "NH":
            hist_sum = hist.sum(dim=-1)  # (B, T)
            hist = hist / hist_sum[:, :, None]
        elif self.hist_type == "LCH":
            hist = torch.log(hist)
        elif self.hist_type != "CH":
            raise ValueError("Invalid value for gateType: gateType should be either IDF or TV")

        return hist

    def _term_gate(self, queries, query_idf, q_masks):
        """
        ： param queries: (B, Tq, H)
        :  param query_idf: (B, Tq)
        ： param q_masks: (B, Tq)
        """
        atten_mask = (1 - q_masks) * -1e7

        if self.gate_type == "IDF":
            gate_prob = self.gates(query_idf[:, :, None]).squeeze() + atten_mask  # (B, 1)
        elif self.gate_type == "TV":
            gate_prob = self.gates(queries).squeeze() + atten_mask  # (B, Tq)
        else:
            raise ValueError("Invalid value for histType: histType should be 'CH', 'NH', or 'LCH'")

        gate_prob = F.softmax(gate_prob, dim=1)  # (B, Tq)
        return gate_prob

    def forward(self, sentence, query_sentence, query_idf):
        query_sent_mask = (query_sentence != 0).to(sentence.device).float()  # (B, Tq)
        sent_mask = (sentence != 0).to(sentence.device).float()  # (B, Td)

        hist_vec = self._hist_map(query_sentence, sentence, sent_mask).to(sentence.device)
        ffw_vec = self.ffw(hist_vec).squeeze().to(sentence.device)  # (B, T1)

        query_idf = query_idf.float()
        query_x = self.embedding(query_sentence).to(sentence.device).float()
        w = self._term_gate(query_x, query_idf, query_sent_mask)  # （B, T1）

        x = (w * ffw_vec).sum(dim=-1, keepdim=True)  # (B, 1)

        score = self.output_layer(x)

        return score


dtype = torch.FloatTensor


@Reranker.register
class DRMM(Reranker):
    """Jiafeng Guo, Yixing Fan, Qingyao Ai, and W. Bruce Croft. 2016. A Deep Relevance Matching Model for Ad-hoc Retrieval. In CIKM'16."""

    module_name = "DRMM"

    config_spec = [
        ConfigOption("nbins", 29, "number of bins in matching histogram"),
        ConfigOption("nodes", 5, "hidden layer dimension for feed forward matching network"),
        ConfigOption("histType", "LCH", "histogram type: CH, NH, or LCH"),
        ConfigOption("gateType", "IDF", "term gate type: TV or IDF"),
    ]

    def build_model(self):
        if not hasattr(self, "model"):
            self.model = DRMM_class(self.extractor, self.config)

        return self.model

    def score(self, d):
        query_idf = d["query_idf"]
        query_sentence = d["query"]
        pos_sentence, neg_sentence = d["posdoc"], d["negdoc"]
        return [
            self.model(pos_sentence, query_sentence, query_idf).view(-1),
            self.model(neg_sentence, query_sentence, query_idf).view(-1),
        ]

    def test(self, d):
        query_idf = d["query_idf"]
        query_sentence = d["query"]
        pos_sentence = d["posdoc"]

        return self.model(pos_sentence, query_sentence, query_idf).view(-1)



import filecmp
import math
import os
import tarfile
from pathlib import Path

import pandas as pd
from tqdm import tqdm

from capreolus import ConfigOption, constants
from capreolus.utils.common import download_file
from capreolus.utils.loginit import get_logger

from . import Collection

logger = get_logger(__name__)
PACKAGE_PATH = constants["PACKAGE_PATH"]


@Collection.register
class COVID(Collection):
    """ The COVID-19 Open Research Dataset (https://www.semanticscholar.org/cord19) """

    module_name = "covid"
    url = "https://ai2-semanticscholar-cord-19.s3-us-west-2.amazonaws.com/historical_releases/cord-19_%s.tar.gz"
    generator_type = "Cord19Generator"
    config_spec = [
        ConfigOption("round", 3),
        ConfigOption("coll_type", "abstract", "one of: abstract, fulltext, paragraph"),
    ]

    def build(self):
        coll_type, round = self.config["coll_type"], self.config["round"]
        type2coll = {
            "abstract": "Cord19AbstractCollection",
            "fulltext": "Cord19FullTextCollection",
            "paragraph": "Cord19ParagraphCollection",
        }
        dates = ["2020-04-10", "2020-05-01", "2020-05-19", "2020-06-19", "2020-07-16"]

        if coll_type not in type2coll:
            raise ValueError(f"Unexpected coll_type: {coll_type}; expeced one of: {' '.join(type2coll.keys())}")
        if round > len(dates):
            raise ValueError(f"Unexpected round number: {round}; only {len(dates)} number of rounds are provided")

        self.collection_type = type2coll[coll_type]
        self.date = dates[round - 1]

    def download_if_missing(self):
        cachedir = self.get_cache_path()
        tmp_dir, document_dir = Path("/tmp"), cachedir / "documents"
        expected_fns = [document_dir / "metadata.csv", document_dir / "document_parses"]
        if all([os.path.exists(f) for f in expected_fns]):
            return document_dir.as_posix()

        url = self.url % self.date
        tar_file = tmp_dir / f"covid-19-{self.date}.tar.gz"
        if not tar_file.exists():
            download_file(url, tar_file)

        with tarfile.open(tar_file) as f:
            f.extractall(path=cachedir)  # emb.tar.gz, metadata.csv, doc.tar.gz, changelog
            os.rename(cachedir / self.date, document_dir)

        doc_fn = "document_parses"
        if f"{doc_fn}.tar.gz" in os.listdir(document_dir):
            with tarfile.open(document_dir / f"{doc_fn}.tar.gz") as f:
                f.extractall(path=document_dir)
        else:
            self.transform_metadata(document_dir)

        # only document_parses and metadata.csv are expected
        for fn in os.listdir(document_dir):
            if (document_dir / fn) not in expected_fns:
                os.remove(document_dir / fn)
        return document_dir.as_posix()

    def transform_metadata(self, root_path):
        """
        the transformation is necessary for dataset round 1 and 2 according to
        https://discourse.cord-19.semanticscholar.org/t/faqs-about-cord-19-dataset/94

        the assumed directory under root_path:
        ./root_path
            ./metadata.csv
            ./comm_use_subset
            ./noncomm_use_subset
            ./custom_license
            ./biorxiv_medrxiv
            ./archive

        In a nutshell:
        1. renaming:
            Microsoft Academic Paper ID -> mag_id;
            WHO #Covidence -> who_covidence_id
        2. update:
            has_pdf_parse -> pdf_json_files  # e.g. document_parses/pmc_json/PMC125340.xml.json
            has_pmc_xml_parse -> pmc_json_files
        """
        metadata_csv = str(root_path / "metadata.csv")
        orifiles = ["arxiv", "custom_license", "biorxiv_medrxiv", "comm_use_subset", "noncomm_use_subset"]
        for fn in orifiles:
            if (root_path / fn).exists():
                continue

            tar_fn = root_path / f"{fn}.tar.gz"
            if not tar_fn.exists():
                continue

            with tarfile.open(str(tar_fn)) as f:
                f.extractall(path=root_path)
                os.remove(tar_fn)

        metadata = pd.read_csv(metadata_csv, header=0)
        columns = metadata.columns.values
        cols_before = [
            "cord_uid",
            "sha",
            "source_x",
            "title",
            "doi",
            "pmcid",
            "pubmed_id",
            "license",
            "abstract",
            "publish_time",
            "authors",
            "journal",
            "Microsoft Academic Paper ID",
            "WHO #Covidence",
            "arxiv_id",
            "has_pdf_parse",
            "has_pmc_xml_parse",
            "full_text_file",
            "url",
        ]
        assert all(columns == cols_before)

        # step 1: rename column
        cols_to_rename = {"Microsoft Academic Paper ID": "mag_id", "WHO #Covidence": "who_covidence_id"}
        metadata.columns = [cols_to_rename.get(c, c) for c in columns]

        # step 2: parse path & move json file
        doc_outp = root_path / "document_parses"
        pdf_dir, pmc_dir = doc_outp / "pdf_json", doc_outp / "pmc_json"
        pdf_dir.mkdir(exist_ok=True, parents=True)
        pmc_dir.mkdir(exist_ok=True, parents=True)

        new_cols = ["pdf_json_files", "pmc_json_files"]
        for col in new_cols:
            metadata[col] = ""
        metadata["s2_id"] = math.nan  # tmp, what's this column??

        iterbar = tqdm(desc="transforming data", total=len(metadata))
        for i, row in metadata.iterrows():
            dir = row["full_text_file"]

            if row["has_pmc_xml_parse"]:
                name = row["pmcid"] + ".xml.json"
                ori_fn = root_path / dir / "pmc_json" / name
                pmc_fn = f"document_parses/pmc_json/{name}"
                metadata.at[i, "pmc_json_files"] = pmc_fn
                pmc_fn = root_path / pmc_fn
                if not pmc_fn.exists():
                    os.rename(ori_fn, pmc_fn)  # check
            else:
                metadata.at[i, "pmc_json_files"] = math.nan

            if row["has_pdf_parse"]:
                shas = str(row["sha"]).split(";")
                pdf_fn_final = ""
                for sha in shas:
                    name = sha.strip() + ".json"
                    ori_fn = root_path / dir / "pdf_json" / name
                    pdf_fn = f"document_parses/pdf_json/{name}"
                    pdf_fn_final = f"{pdf_fn_final};{pdf_fn}" if pdf_fn_final else pdf_fn
                    pdf_fn = root_path / pdf_fn
                    if not pdf_fn.exists():
                        os.rename(ori_fn, pdf_fn)  # check
                    else:
                        if ori_fn.exists():
                            assert filecmp.cmp(ori_fn, pdf_fn)
                            os.remove(ori_fn)

                metadata.at[i, "pdf_json_files"] = pdf_fn_final
            else:
                metadata.at[i, "pdf_json_files"] = math.nan

            iterbar.update()

        # step 3: remove deprecated columns, remove unwanted directories
        cols_to_remove = ["has_pdf_parse", "has_pmc_xml_parse", "full_text_file"]
        metadata.drop(columns=cols_to_remove)

        dir_to_remove = ["comm_use_subset", "noncomm_use_subset", "custom_license", "biorxiv_medrxiv", "arxiv"]
        for dir in dir_to_remove:
            dir = root_path / dir
            for subdir in os.listdir(dir):
                os.rmdir(dir / subdir)  # since we are supposed to move away all the files
            os.rmdir(dir)

        # assert len(metadata.columns) == 19
        # step 4: save back
        metadata.to_csv(metadata_csv, index=False)



import copy

import torch
import torch.nn.functional as F
from torch import nn
from torch.autograd import Variable

from capreolus import ConfigOption, Dependency
from capreolus.reranker import Reranker
from capreolus.utils.loginit import get_logger

logger = get_logger(__name__)  # pylint: disable=invalid-name


class DeepTileBar_nn(nn.Module):
    def __init__(self, p, batch_size, number_filter, lstm_hidden_dim, linear_hidden_dim1, linear_hidden_dim2):
        super(DeepTileBar_nn, self).__init__()
        self.p = p
        self.tilechannels = 3
        if not self.p["tfchannel"]:
            self.tilechannels -= 1
        self.batch_size = batch_size
        self.number_filter = number_filter
        self.lstm_hidden_dim = lstm_hidden_dim
        self.conv1 = nn.Conv2d(self.tilechannels, number_filter, (p["maxqlen"], 1), stride=1)
        self.conv2 = nn.Conv2d(self.tilechannels, number_filter, (p["maxqlen"], 2), stride=1)
        self.conv3 = nn.Conv2d(self.tilechannels, number_filter, (p["maxqlen"], 3), stride=1)
        self.conv4 = nn.Conv2d(self.tilechannels, number_filter, (p["maxqlen"], 4), stride=1)
        self.conv5 = nn.Conv2d(self.tilechannels, number_filter, (p["maxqlen"], 5), stride=1)
        self.conv6 = nn.Conv2d(self.tilechannels, number_filter, (p["maxqlen"], 6), stride=1)
        self.conv7 = nn.Conv2d(self.tilechannels, number_filter, (p["maxqlen"], 7), stride=1)
        self.conv8 = nn.Conv2d(self.tilechannels, number_filter, (p["maxqlen"], 8), stride=1)
        self.conv9 = nn.Conv2d(self.tilechannels, number_filter, (p["maxqlen"], 9), stride=1)
        self.conv10 = nn.Conv2d(self.tilechannels, number_filter, (p["maxqlen"], 10), stride=1)
        self.lstm1 = nn.LSTM(input_size=3, hidden_size=lstm_hidden_dim)
        self.lstm2 = nn.LSTM(input_size=3, hidden_size=lstm_hidden_dim)
        self.lstm3 = nn.LSTM(input_size=3, hidden_size=lstm_hidden_dim)
        self.lstm4 = nn.LSTM(input_size=3, hidden_size=lstm_hidden_dim)
        self.lstm5 = nn.LSTM(input_size=3, hidden_size=lstm_hidden_dim)
        self.lstm6 = nn.LSTM(input_size=3, hidden_size=lstm_hidden_dim)
        self.lstm7 = nn.LSTM(input_size=3, hidden_size=lstm_hidden_dim)
        self.lstm8 = nn.LSTM(input_size=3, hidden_size=lstm_hidden_dim)
        self.lstm9 = nn.LSTM(input_size=3, hidden_size=lstm_hidden_dim)
        self.lstm10 = nn.LSTM(input_size=3, hidden_size=lstm_hidden_dim)
        self.W1 = nn.Linear(10 * lstm_hidden_dim, linear_hidden_dim1, bias=True)
        self.W2 = nn.Linear(linear_hidden_dim1, linear_hidden_dim2, bias=True)
        self.W3 = nn.Linear(linear_hidden_dim2, 1, bias=True)
        [
            self.hidden1,
            self.hidden2,
            self.hidden3,
            self.hidden4,
            self.hidden5,
            self.hidden6,
            self.hidden7,
            self.hidden8,
            self.hidden9,
            self.hidden10,
        ] = self.init_hidden()

    def init_hidden(self):
        # first is the hidden h
        # second is the cell c
        # if self.use_gpu:
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        l = []
        for j in range(10):
            l.append(
                (
                    Variable(torch.zeros(1, self.batch_size, self.lstm_hidden_dim).to(device)),
                    Variable(torch.zeros(1, self.batch_size, self.lstm_hidden_dim).to(device)),
                )
            )
        return l

    def reset_hidden(self):
        [
            self.hidden1,
            self.hidden2,
            self.hidden3,
            self.hidden4,
            self.hidden5,
            self.hidden6,
            self.hidden7,
            self.hidden8,
            self.hidden9,
            self.hidden10,
        ] = self.init_hidden()

    def forward(self, tile_matrix1):
        tile_matrix2 = torch.transpose(
            torch.transpose(tile_matrix1.view(self.batch_size, self.p["maxqlen"], self.p["passagelen"], -1), 1, 3), 2, 3
        )
        x1 = torch.transpose(torch.transpose(self.conv1(tile_matrix2).view(self.batch_size, self.number_filter, -1), 0, 2), 1, 2)
        x2 = torch.transpose(torch.transpose(self.conv2(tile_matrix2).view(self.batch_size, self.number_filter, -1), 0, 2), 1, 2)
        x3 = torch.transpose(torch.transpose(self.conv3(tile_matrix2).view(self.batch_size, self.number_filter, -1), 0, 2), 1, 2)
        x4 = torch.transpose(torch.transpose(self.conv4(tile_matrix2).view(self.batch_size, self.number_filter, -1), 0, 2), 1, 2)
        x5 = torch.transpose(torch.transpose(self.conv5(tile_matrix2).view(self.batch_size, self.number_filter, -1), 0, 2), 1, 2)
        x6 = torch.transpose(torch.transpose(self.conv6(tile_matrix2).view(self.batch_size, self.number_filter, -1), 0, 2), 1, 2)
        x7 = torch.transpose(torch.transpose(self.conv7(tile_matrix2).view(self.batch_size, self.number_filter, -1), 0, 2), 1, 2)
        x8 = torch.transpose(torch.transpose(self.conv8(tile_matrix2).view(self.batch_size, self.number_filter, -1), 0, 2), 1, 2)
        x9 = torch.transpose(torch.transpose(self.conv9(tile_matrix2).view(self.batch_size, self.number_filter, -1), 0, 2), 1, 2)
        x10 = torch.transpose(
            torch.transpose(self.conv10(tile_matrix2).view(self.batch_size, self.number_filter, -1), 0, 2), 1, 2
        )

        lstm_out1, self.hidden1 = self.lstm1(x1, self.hidden1)
        lstm_out2, self.hidden2 = self.lstm2(x2, self.hidden2)
        lstm_out3, self.hidden3 = self.lstm3(x3, self.hidden3)
        lstm_out4, self.hidden4 = self.lstm4(x4, self.hidden4)
        lstm_out5, self.hidden5 = self.lstm5(x5, self.hidden5)
        lstm_out6, self.hidden6 = self.lstm6(x6, self.hidden6)
        lstm_out7, self.hidden7 = self.lstm7(x7, self.hidden7)
        lstm_out8, self.hidden8 = self.lstm8(x8, self.hidden8)
        lstm_out9, self.hidden9 = self.lstm9(x9, self.hidden9)
        lstm_out10, self.hidden10 = self.lstm10(x10, self.hidden10)
        input_x = torch.cat(
            [
                lstm_out1[-1],
                lstm_out2[-1],
                lstm_out3[-1],
                lstm_out4[-1],
                lstm_out5[-1],
                lstm_out6[-1],
                lstm_out7[-1],
                lstm_out8[-1],
                lstm_out9[-1],
                lstm_out10[-1],
            ],
            1,
        )
        input_x1 = F.relu(self.W1(input_x))
        input_x2 = F.relu(self.W2(input_x1))
        input_x3 = self.W3(input_x2)
        return input_x3.view(-1)


class DeepTileBar_class(nn.Module):
    def __init__(self, extractor, config):
        super(DeepTileBar_class, self).__init__()
        batch_size = config["batch"]
        number_filter = config["numberfilter"]
        lstm_hidden_dim = config["lstmhiddendim"]
        linear_hidden_dim1 = config["linearhiddendim1"]
        linear_hidden_dim2 = config["linearhiddendim2"]
        config = dict(config)
        config.update(dict(extractor.config))

        self.DeepTileBar1 = DeepTileBar_nn(
            config, batch_size, number_filter, lstm_hidden_dim, linear_hidden_dim1, linear_hidden_dim2
        )

    def init_hidden(self):
        return self.DeepTileBar1.init_hidden()

    def reset_hidden(self):
        self.DeepTileBar1.reset_hidden()

    def forward(self, pos_tile_matrix, neg_tile_matrix):
        self.reset_hidden()
        pos_tag_scores = self.DeepTileBar1(pos_tile_matrix)
        self.reset_hidden()
        neg_tag_scores = self.DeepTileBar1(neg_tile_matrix)
        return [pos_tag_scores, neg_tag_scores]

    def test_forward(self, pos_tile_matrix):
        self.reset_hidden()
        pos_tag_scores = self.DeepTileBar1(pos_tile_matrix)
        return pos_tag_scores


@Reranker.register
class DeepTileBar(Reranker):
    """Zhiwen Tang and Grace Hui Yang. 2019. DeepTileBars: Visualizing Term Distribution for Neural Information Retrieval. In AAAI'19."""

    module_name = "DeepTileBar"

    dependencies = [
        Dependency(key="extractor", module="extractor", name="deeptiles"),
        Dependency(key="trainer", module="trainer", name="pytorch"),
    ]

    config_spec = [
        ConfigOption("passagelen", 30),
        ConfigOption("numberfilter", 3),
        ConfigOption("lstmhiddendim", 3),
        ConfigOption("linearhiddendim1", 32),
        ConfigOption("linearhiddendim2", 16),
    ]

    def build_model(self):
        if not hasattr(self, "model"):
            config = copy.copy(dict(self.config))
            config["batch"] = self.trainer.config["batch"]
            self.model = DeepTileBar_class(self.extractor, config)

        return self.model

    def score(self, d):
        pos_tile_matrix = torch.cat([d["posdoc"][i] for i in range(len(d["qid"]))])  # 32 x
        neg_tile_matrix = torch.cat([d["negdoc"][i] for i in range(len(d["qid"]))])
        return self.model(pos_tile_matrix, neg_tile_matrix)

    def test(self, d):
        qids = d["qid"]
        pos_sentence = d["posdoc"]
        pos_tile_matrix = torch.cat([pos_sentence[i] for i in range(len(qids))])

        return self.model.test_forward(pos_tile_matrix)

    def zero_grad(self, *args, **kwargs):
        self.model.zero_grad(*args, **kwargs)



import os

from capreolus import ModuleBase


class Collection(ModuleBase):
    """Base class for Collection modules. The purpose of a Collection is to describe a document collection's location and its format.

    Determining the document collection's location on disk:
        - The *path* config option will be used if it contains a valid loation.
        - If not, the ``_path`` attribute is used if it is valid. This is primarily used with :class:`~.dummy.DummyCollection`.
        - If not, the class' ``download_if_missing`` method will be called.

    Modules should provide:
        - the ``collection_type`` and ``generator_type`` class attributes, corresponding to Anserini types
        - a ``download_if_missing`` method, if the collection is publicly available
        - a ``_validate_document_path`` method. See :func:`~capreolus.collection.Collection.validate_document_path`.
    """

    module_type = "collection"
    is_large_collection = False
    _path = None

    def __iter__(self):
        from pyserini.collection import pycollection
        from pyserini.index import pygenerator

        path, ctype, gtype = self.get_path_and_types()
        # TODO change on pyserini upgrade
        if gtype == "WashingtonPostGenerator":
            gtype = "WapoGenerator"

        collection = pycollection.Collection(ctype, path)
        generator = pygenerator.Generator(gtype)

        for fs in collection:
            for doc in fs:
                parsed = None
                try:
                    parsed = generator.create_document(doc)
                except:
                    pass

                if parsed:
                    yield (parsed.get("id"), parsed.get("title"), parsed.get("contents"))

    def get_path_and_types(self):
        """ Returns a ``(path, collection_type, generator_type)`` tuple. """
        if not self.validate_document_path(self._path):
            self._path = self.find_document_path()

        return self._path, self.collection_type, self.generator_type

    def validate_document_path(self, path):
        """Attempt to validate the document collection at ``path``.

        By default, this will only check whether ``path`` exists. Subclasses should override
        ``_validate_document_path(path)`` with their own logic to perform more detailed checks.

        Returns:
            True if the path is valid following the logic described above, or False if it is not
        """

        if not (path and os.path.exists(path)):
            return False

        return self._validate_document_path(path)

    def _validate_document_path(self, path):
        """Collection-specific logic for validating the document collection path. Subclasses should override this.

        Returns:
            this default method provided by Collection always returns true
        """

        return True

    def find_document_path(self):
        """Find the location of this collection's documents (i.e., the raw document collection).

        We first check the collection's config for a path key. If found, ``self.validate_document_path`` checks
        whether the path is valid. Subclasses should override the private method ``self._validate_document_path``
        with custom logic for performing checks further than existence of the directory.

        If a valid path was not found, call ``download_if_missing``.
        Subclasses should override this method if downloading the needed documents is possible.

        If a valid document path cannot be found, an exception is thrown.

        Returns:
            path to this collection's raw documents
        """

        # first, see if the path was provided as a config option
        if "path" in self.config and self.validate_document_path(self.config["path"]):
            return self.config["path"]

        # see if the path is hardcoded (e.g., for the dummy collection")
        if self._path and self.validate_document_path(self._path):
            return self._path

        # if not, see if the collection can be obtained through its download_if_missing method
        return self.download_if_missing()

    def download_if_missing(self):
        """ Download the collection and return its path. Subclasses should override this. """
        raise IOError(
            f"a download URL is not configured for collection={self.module_name} and the collection path does not exist; you must manually place the document collection at this path in order to use this collection"
        )


from profane import import_all_modules

from .dummy import DummyCollection
from .antique import ANTIQUE
from .nf import NF
from .robust04 import Robust04

import_all_modules(__file__, __package__)



import tensorflow as tf
from transformers import TFBertModel, TFElectraModel
from transformers.modeling_tf_bert import TFBertLayer

from capreolus import ConfigOption, Dependency
from capreolus.reranker import Reranker


class TFParade_Class(tf.keras.layers.Layer):
    def __init__(self, extractor, config, *args, **kwargs):
        super(TFParade_Class, self).__init__(*args, **kwargs)
        self.extractor = extractor
        self.config = config

        if config["pretrained"] == "electra-base-msmarco":
            self.bert = TFElectraModel.from_pretrained("Capreolus/electra-base-msmarco")
        elif config["pretrained"] == "bert-base-msmarco":
            self.bert = TFBertModel.from_pretrained("Capreolus/bert-base-msmarco")
        elif config["pretrained"] == "bert-base-uncased":
            self.bert = TFBertModel.from_pretrained("bert-base-uncased")
        else:
            raise ValueError(
                f"unsupported model: {config['pretrained']}; need to ensure correct tokenizers will be used before arbitrary hgf models are supported"
            )

        self.transformer_layer_1 = TFBertLayer(self.bert.config)
        self.transformer_layer_2 = TFBertLayer(self.bert.config)
        self.num_passages = extractor.config["numpassages"]
        self.maxseqlen = extractor.config["maxseqlen"]
        self.linear = tf.keras.layers.Dense(1, input_shape=(self.bert.config.hidden_size,))

        if config["aggregation"] == "maxp":
            self.aggregation = self.aggregate_using_maxp
        elif config["aggregation"] == "transformer":
            self.aggregation = self.aggregate_using_transformer
            input_embeddings = self.bert.get_input_embeddings()
            cls_token_id = tf.convert_to_tensor([101])
            cls_token_id = tf.reshape(cls_token_id, [1, 1])
            self.initial_cls_embedding = input_embeddings(input_ids=cls_token_id)
            self.initial_cls_embedding = tf.reshape(self.initial_cls_embedding, [1, self.bert.config.hidden_size])
            initializer = tf.random_normal_initializer(stddev=0.02)
            full_position_embeddings = tf.Variable(
                initial_value=initializer(shape=[self.num_passages + 1, self.bert.config.hidden_size]),
                name="passage_position_embedding",
            )
            self.full_position_embeddings = tf.expand_dims(full_position_embeddings, axis=0)

    def aggregate_using_maxp(self, cls):
        """
        cls has the shape [B, num_passages, hidden_size]
        """
        expanded_cls = tf.reshape(cls, [-1, self.num_passages, self.bert.config.hidden_size])
        aggregated = tf.reduce_max(expanded_cls, axis=1)

        return aggregated

    def aggregate_using_transformer(self, cls):
        expanded_cls = tf.reshape(cls, [-1, self.num_passages, self.bert.config.hidden_size])
        batch_size = tf.shape(expanded_cls)[0]
        tiled_initial_cls = tf.tile(self.initial_cls_embedding, multiples=[batch_size, 1])
        merged_cls = tf.concat((tf.expand_dims(tiled_initial_cls, axis=1), expanded_cls), axis=1)

        merged_cls += self.full_position_embeddings

        (transformer_out_1,) = self.transformer_layer_1(merged_cls, None, None, None)
        (transformer_out_2,) = self.transformer_layer_2(transformer_out_1, None, None, None)

        aggregated = transformer_out_2[:, 0, :]
        return aggregated

    def call(self, x, **kwargs):
        doc_input, doc_mask, doc_seg = x[0], x[1], x[2]
        batch_size = tf.shape(doc_input)[0]

        doc_input = tf.reshape(doc_input, [batch_size * self.num_passages, self.maxseqlen])
        doc_mask = tf.reshape(doc_mask, [batch_size * self.num_passages, self.maxseqlen])
        doc_seg = tf.reshape(doc_seg, [batch_size * self.num_passages, self.maxseqlen])

        cls = self.bert(doc_input, attention_mask=doc_mask, token_type_ids=doc_seg)[0][:, 0, :]
        aggregated = self.aggregation(cls)

        return self.linear(aggregated)

    def predict_step(self, data):
        """
        Scores each passage and applies max pooling over it.
        """
        posdoc_bert_input, posdoc_mask, posdoc_seg, negdoc_bert_input, negdoc_mask, negdoc_seg = data
        doc_scores = self.call((posdoc_bert_input, posdoc_mask, posdoc_seg), training=False)

        return doc_scores

    def score(self, x, **kwargs):
        posdoc_bert_input, posdoc_mask, posdoc_seg, negdoc_bert_input, negdoc_mask, negdoc_seg = x

        return self.call((posdoc_bert_input, posdoc_mask, posdoc_seg), **kwargs)

    def score_pair(self, x, **kwargs):
        posdoc_bert_input, posdoc_mask, posdoc_seg, negdoc_bert_input, negdoc_mask, negdoc_seg = x

        pos_score = self.call((posdoc_bert_input, posdoc_mask, posdoc_seg), **kwargs)
        neg_score = self.call((negdoc_bert_input, negdoc_mask, negdoc_seg), **kwargs)

        return pos_score, neg_score


@Reranker.register
class TFParade(Reranker):
    module_name = "parade"

    dependencies = [
        Dependency(key="extractor", module="extractor", name="pooledbertpassage"),
        Dependency(key="trainer", module="trainer", name="tensorflow"),
    ]
    config_spec = [
        ConfigOption(
            "pretrained", "bert-base-uncased", "Pretrained model: bert-base-uncased, bert-base-msmarco, or electra-base-msmarco"
        ),
        ConfigOption("aggregation", "transformer"),
    ]

    def build_model(self):
        self.model = TFParade_Class(self.extractor, self.config)
        return self.model



from capreolus import Dependency, constants
from capreolus.utils.loginit import get_logger

from . import Benchmark

logger = get_logger(__name__)
PACKAGE_PATH = constants["PACKAGE_PATH"]


@Benchmark.register
class ANTIQUE(Benchmark):
    """A Non-factoid Question Answering Benchmark from Hashemi et al. [1]

    [1] Helia Hashemi, Mohammad Aliannejadi, Hamed Zamani, and W. Bruce Croft. 2020. ANTIQUE: A non-factoid question answering benchmark. ECIR 2020.
    """

    module_name = "antique"
    dependencies = [Dependency(key="collection", module="collection", name="antique")]
    qrel_file = PACKAGE_PATH / "data" / "qrels.antique.txt"
    topic_file = PACKAGE_PATH / "data" / "topics.antique.txt"
    fold_file = PACKAGE_PATH / "data" / "antique.json"
    query_type = "title"
    relevance_level = 2



import torch
from torch import nn

from capreolus import ConfigOption, Dependency
from capreolus.reranker import Reranker
from capreolus.reranker.common import RbfKernelBank, StackedSimilarityMatrix, create_emb_layer
from capreolus.utils.loginit import get_logger

logger = get_logger(__name__)  # pylint: disable=invalid-name


class ConvKNRM_class(nn.Module):
    def __init__(self, extractor, config):
        super(ConvKNRM_class, self).__init__()
        self.p = config
        self.simmat = StackedSimilarityMatrix(padding=extractor.pad)
        self.embeddings = create_emb_layer(extractor.embeddings, non_trainable=True)

        mus = [-0.9, -0.7, -0.5, -0.3, -0.1, 0.1, 0.3, 0.5, 0.7, 0.9, 1.0]
        sigmas = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.001]
        self.kernels = RbfKernelBank(mus, sigmas, dim=1, requires_grad=config["gradkernels"])

        self.padding, self.convs = nn.ModuleList(), nn.ModuleList()
        for conv_size in range(1, config["maxngram"] + 1):
            if conv_size > 1:
                self.padding.append(nn.ConstantPad1d((0, conv_size - 1), 0))
            else:
                self.padding.append(nn.Sequential())  # identity
            self.convs.append(nn.ModuleList())
            for _ in range(1):
                self.convs[-1].append(nn.Conv1d(self.embeddings.weight.shape[1], config["filters"], conv_size))

        channels = config["maxngram"] ** 2 if config["crossmatch"] else config["maxngram"]
        channels *= 1 ** 2 if config["crossmatch"] else 1
        if config["singlefc"]:
            combine_steps = [nn.Linear(self.kernels.count() * channels, 1)]
        else:
            combine_steps = [nn.Linear(self.kernels.count() * channels, 30), nn.Tanh(), nn.Linear(30, 1)]
        if config["scoretanh"]:
            combine_steps.append(nn.Tanh())
        self.combine = nn.Sequential(*combine_steps)

    def forward(self, sentence, query_sentence, query_idf):
        a_embed = [self.embeddings(query_sentence)]
        b_embed = [self.embeddings(sentence)]
        a_reps, b_reps = [], []
        for layer, (a_emb, b_emb) in enumerate(zip(a_embed, b_embed)):
            for pad, conv in zip(self.padding, self.convs):
                a_reps.append(conv[layer](pad(a_emb.permute(0, 2, 1))).permute(0, 2, 1))
                b_reps.append(conv[layer](pad(b_emb.permute(0, 2, 1))).permute(0, 2, 1))

        simmats = []
        if self.p["crossmatch"]:
            for a_rep in a_reps:
                for b_rep in b_reps:
                    simmats.append(self.simmat(a_rep, b_rep, query_sentence, sentence))
        else:
            for i, a_rep in enumerate(a_reps):
                b_rep = b_reps[i]
                simmats.append(self.simmat(a_rep, b_rep, query_sentence, sentence))
        simmats = torch.cat(simmats, dim=1)

        # remainder is the same as KNRM
        kernels = self.kernels(simmats)
        BATCH, KERNELS, VIEWS, QLEN, DLEN = kernels.shape
        kernels = kernels.reshape(BATCH, KERNELS * VIEWS, QLEN, DLEN)
        simmats = (
            simmats.reshape(BATCH, 1, VIEWS, QLEN, DLEN)
            .expand(BATCH, KERNELS, VIEWS, QLEN, DLEN)
            .reshape(BATCH, KERNELS * VIEWS, QLEN, DLEN)
        )
        result = kernels.sum(dim=3)  # sum over document
        mask = simmats.sum(dim=3) != 0.0  # which query terms are not padding?
        result = torch.where(mask, (result + 1e-6).log(), mask.float())
        result = result.sum(dim=2)  # sum over query terms
        scores = self.combine(result)  # linear combination over kernels
        return scores


@Reranker.register
class ConvKNRM(Reranker):
    """Zhuyun Dai, Chenyan Xiong, Jamie Callan, and Zhiyuan Liu. 2018. Convolutional Neural Networks for Soft-Matching N-Grams in Ad-hoc Search. In WSDM'18."""

    module_name = "ConvKNRM"

    dependencies = [
        Dependency(key="extractor", module="extractor", name="slowembedtext"),
        Dependency(key="trainer", module="trainer", name="pytorch"),
    ]
    config_spec = [
        ConfigOption("gradkernels", True, "backprop through mus and sigmas"),
        ConfigOption("maxngram", 3, "maximum ngram length considered"),
        ConfigOption("crossmatch", True, "match query and document ngrams of different lengths (e.g., bigram vs. unigram)"),
        ConfigOption("filters", 128, "number of filters used in convolutional layers"),
        ConfigOption("scoretanh", False, "use a tanh on the prediction as in paper (True) or do not use a nonlinearity (False)"),
        ConfigOption("singlefc", True, "use single fully connected layer as in paper (True) or 2 fully connected layers (False)"),
    ]

    def build_model(self):
        if not hasattr(self, "model"):
            self.model = ConvKNRM_class(self.extractor, self.config)
        return self.model

    def score(self, d):
        query_idf = d["query_idf"]
        query_sentence = d["query"]
        pos_sentence, neg_sentence = d["posdoc"], d["negdoc"]
        return [
            self.model(pos_sentence, query_sentence, query_idf).view(-1),
            self.model(neg_sentence, query_sentence, query_idf).view(-1),
        ]

    def test(self, d):
        query_idf = d["query_idf"]
        query_sentence = d["query"]
        pos_sentence = d["posdoc"]
        return self.model(pos_sentence, query_sentence, query_idf).view(-1)

    def zero_grad(self, *args, **kwargs):
        self.model.zero_grad(*args, **kwargs)



import json

from capreolus import ModuleBase
from capreolus.utils.trec import load_qrels, load_trec_topics


class Benchmark(ModuleBase):
    """Base class for Benchmark modules. The purpose of a Benchmark is to provide the data needed to run an experiment, such as queries, folds, and relevance judgments.

    Modules should provide:
        - a ``topics`` dict mapping query ids (*qids*) to *queries*
        - a ``qrels`` dict mapping *qids* to *docids* and *relevance labels*
        - a ``folds`` dict mapping a fold name to *training*, *dev* (validation), and *testing* qids
        - if these can be loaded from files in standard formats, they can be specified by setting the ``topic_file``, ``qrel_file``, and ``fold_file``, respectively, rather than by setting the above attributes directly
    """

    module_type = "benchmark"
    qrel_file = None
    topic_file = None
    fold_file = None
    query_type = None
    relevance_level = 1
    """ Documents with a relevance label >= relevance_level will be considered relevant.
    This corresponds to trec_eval's --level_for_rel (and is passed to pytrec_eval as relevance_level). """

    @property
    def qrels(self):
        if not hasattr(self, "_qrels"):
            self._qrels = load_qrels(self.qrel_file)
        return self._qrels

    @property
    def topics(self):
        if not hasattr(self, "_topics"):
            self._topics = load_trec_topics(self.topic_file)
        return self._topics

    @property
    def folds(self):
        if not hasattr(self, "_folds"):
            self._folds = json.load(open(self.fold_file, "rt"), parse_int=str)
        return self._folds


from profane import import_all_modules

from .dummy import DummyBenchmark

import_all_modules(__file__, __package__)



import torch
import torch.nn as nn

from capreolus import ConfigOption, Dependency
from capreolus.reranker import Reranker
from capreolus.utils.loginit import get_logger

logger = get_logger(__name__)


class DSSM_class(nn.Module):
    def __init__(self, extractor, config):
        super(DSSM_class, self).__init__()
        p = config
        nvocab = len(extractor.stoi)
        nhiddens = [nvocab] + list(map(int, p["nhiddens"].split()))
        self.ffw = nn.Sequential()
        for i in range(len(nhiddens) - 1):
            self.ffw.add_module("linear%d" % i, nn.Linear(nhiddens[i], nhiddens[i + 1]))
            self.ffw.add_module("activate%d" % i, nn.ReLU())
            self.ffw.add_module("dropout%i" % i, nn.Dropout(0.5))

        self.output_layer = nn.Sigmoid()

    def forward(self, sentence, query, query_idf):
        query = query.float()
        sentence = sentence.float()
        query = self.ffw(query)
        sentence = self.ffw(sentence)

        query_norm = query.norm(dim=-1)[:, None] + 1e-7
        sentence_norm = sentence.norm(dim=-1)[:, None] + 1e-7

        query = query / query_norm
        sentence = sentence / sentence_norm

        cos_x = (query * sentence).sum(dim=-1, keepdim=True)

        score = self.output_layer(cos_x)
        return score


dtype = torch.FloatTensor


@Reranker.register
class DSSM(Reranker):
    """Po-Sen Huang, Xiaodong He, Jianfeng Gao, Li Deng, Alex Acero, and Larry Heck. 2013. Learning deep structured semantic models for web search using clickthrough data. In CIKM'13."""

    module_name = "DSSM"
    dependencies = [
        Dependency(key="extractor", module="extractor", name="bagofwords"),
        Dependency(key="trainer", module="trainer", name="pytorch", default_config_overrides={"lr": 0.0001}),
    ]
    config_spec = [
        ConfigOption(
            "nhiddens",
            "56",
            "list of hidden layer sizes (eg '56 128'), where the i'th value indicates the output size of the i'th layer",
        )
    ]

    def build_model(self):
        if not hasattr(self, "model"):
            self.model = DSSM_class(self.extractor, self.config)
        return self.model

    def score(self, d):
        query_idf = d["query_idf"]
        query_sentence = d["query"]
        pos_sentence, neg_sentence = d["posdoc"], d["negdoc"]
        return [
            self.model(pos_sentence, query_sentence, query_idf).view(-1),
            self.model(neg_sentence, query_sentence, query_idf).view(-1),
        ]

    def test(self, d):
        query_idf = d["query_idf"]
        query_sentence = d["query"]
        pos_sentence = d["posdoc"]
        return self.model(pos_sentence, query_sentence, query_idf).view(-1)



from capreolus import Dependency, constants

from . import Benchmark

PACKAGE_PATH = constants["PACKAGE_PATH"]


@Benchmark.register
class Robust04(Benchmark):
    """Robust04 benchmark using the title folds from Huston and Croft. [1] Each of these is used as the test set.
    Given the remaining four folds, we split them into the same train and dev sets used in recent work. [2]

    [1] Samuel Huston and W. Bruce Croft. 2014. Parameters learned in the comparison of retrieval models using term dependencies. Technical Report.

    [2] Sean MacAvaney, Andrew Yates, Arman Cohan, Nazli Goharian. 2019. CEDR: Contextualized Embeddings for Document Ranking. SIGIR 2019.
    """

    module_name = "robust04"
    dependencies = [Dependency(key="collection", module="collection", name="robust04")]
    qrel_file = PACKAGE_PATH / "data" / "qrels.robust2004.txt"
    topic_file = PACKAGE_PATH / "data" / "topics.robust04.301-450.601-700.txt"
    fold_file = PACKAGE_PATH / "data" / "rob04_cedr_folds.json"
    query_type = "title"


@Benchmark.register
class Robust04Yang19(Benchmark):
    """Robust04 benchmark using the folds from Yang et al. [1]

    [1] Wei Yang, Kuang Lu, Peilin Yang, and Jimmy Lin. 2019. Critically Examining the "Neural Hype": Weak Baselines and the Additivity of Effectiveness Gains from Neural Ranking Models. SIGIR 2019.
    """

    module_name = "robust04.yang19"
    dependencies = [Dependency(key="collection", module="collection", name="robust04")]
    qrel_file = PACKAGE_PATH / "data" / "qrels.robust2004.txt"
    topic_file = PACKAGE_PATH / "data" / "topics.robust04.301-450.601-700.txt"
    fold_file = PACKAGE_PATH / "data" / "rob04_yang19_folds.json"
    query_type = "title"



import math

import torch
import torch.nn.functional as F
from torch import nn
from torch.autograd import Variable

from capreolus import ConfigOption, Dependency
from capreolus.reranker import Reranker
from capreolus.reranker.common import create_emb_layer

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


class GRUCell2d(nn.Module):
    def __init__(self, input_size, hidden_size, bias=True):
        super(GRUCell2d, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.bias = bias
        self.Wrz = nn.Linear(3 * hidden_size + input_size, 7 * hidden_size, bias=bias)
        self.W = nn.Linear(input_size, hidden_size, bias=bias)
        self.U = nn.Linear(3 * hidden_size, hidden_size, bias=bias)

    def reset_parameters(self):
        std = 1.0 / math.sqrt(self.hidden_size)
        for w in self.parameters():
            w.data.uniform_(-std, std)

    def softmax_by_row(self, zi, zl, zt, zd):
        zi, zl, zt, zd = zi.view(-1, 1, 2), zl.view(-1, 1, 2), zt.view(-1, 1, 2), zd.view(-1, 1, 2)
        # each: (B, 1, hidden=2)

        ppp = torch.cat((zi, zl, zt, zd), dim=1)  # (B, 4, hidden)

        pt = F.softmax(ppp, dim=1)  # (B, 4, hidden)

        zi, zl, zt, zd = pt.unbind(dim=1)
        zi, zl, zt, zd = zi.view(-1, 2), zl.view(-1, 2), zt.view(-1, 2), zd.view(-1, 2)

        return zi, zl, zt, zd

    def forward(self, x, hidden_i1_j1, hidden_i1_j, hidden_i_j1):
        q = torch.cat([hidden_i1_j, hidden_i_j1, hidden_i1_j1, x], dim=-1)  # (B, 3*nhidden+input)
        r_z = self.Wrz(q)
        rl, rt, rd, zi, zl, zt, zd = r_z.chunk(7, 1)  # each: (B, hidden)

        rl, rt, rd = torch.sigmoid(rl), torch.sigmoid(rt), torch.sigmoid(rd)
        zi, zl, zt, zd = self.softmax_by_row(zi, zl, zt, zd)

        r = torch.cat([rl, rt, rd], dim=-1)
        t11 = torch.cat([hidden_i1_j, hidden_i_j1, hidden_i1_j1], dim=-1)

        h1 = torch.tanh(self.W(x) + self.U(r * t11))
        h = (zl * hidden_i_j1) + (zt * hidden_i1_j) + (zd * hidden_i1_j1) + (zi * h1)

        return h


class GRUModel2d(nn.Module):
    def __init__(self, input_dim, hidden_dim, bias=True):
        super(GRUModel2d, self).__init__()
        self.hidden_dim = hidden_dim
        self.gru_cell = GRUCell2d(input_dim, hidden_dim).to(device)

    def forward(self, x):
        B, T1, T2, H = x.size()
        last_outs = [(torch.zeros(x.size(0), self.hidden_dim).to(device)) for _ in range(T2 + 1)]
        for seq in range(T1):
            outs_row = [(torch.zeros(x.size(0), self.hidden_dim).to(device))]
            for seq1 in range(1, T2 + 1):
                hn = last_outs[seq1 - 1]
                hn_top = last_outs[seq1]
                hn_left = outs_row[seq1 - 1]

                hn1 = self.gru_cell(x[:, seq, seq1 - 1, :], hn, hn_top, hn_left)
                outs_row.append(hn1)

            last_outs = outs_row
        out = last_outs[-1]
        return out


class HiNT(nn.Module):
    def __init__(self, weights_matrix, p):
        super(HiNT, self).__init__()
        self.p = p
        self.passagelen = int(p["maxdoclen"] / 100)  # 100: windows size
        Ws_dim = 1  # fix to 1, since we assume 1 when creating GRUModel (2*Ws_dim + 1)

        embedding_dim = weights_matrix.shape[1]
        self.batch_size, self.lstm_hidden_dim = p["batch"], self.p["LSTMdim"]

        self.embedding = create_emb_layer(weights_matrix, non_trainable=True)

        self.Ws = nn.Linear(embedding_dim, Ws_dim)
        self.GRU2d1 = GRUModel2d(3, self.p["spatialGRU"]).to(device)
        self.GRU2d3 = GRUModel2d(3, self.p["spatialGRU"]).to(device)

        self.lstm = nn.LSTM(input_size=(4 * self.p["spatialGRU"]), hidden_size=self.lstm_hidden_dim, bidirectional=True)
        self.Wv = nn.Linear((4 * self.p["spatialGRU"]), self.lstm_hidden_dim, bias=True)
        self.fc = nn.Linear(self.lstm_hidden_dim * self.p["kmax"], 1)
        self.hidden = self.init_hidden()

    def init_hidden(self):
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        return (
            Variable(torch.zeros(2, self.batch_size, self.lstm_hidden_dim).to(device)),
            Variable(torch.zeros(2, self.batch_size, self.lstm_hidden_dim).to(device)),
        )

    def matrix_inv(self, A):
        A1 = torch.randn(self.passagelen * self.batch_size, self.p["maxqlen"], 100, 3).type(torch.FloatTensor).to(device)
        for i in range(self.p["maxqlen"]):
            for j in range(100):
                A1[:, i, j, :] = A[:, self.p["maxqlen"] - i - 1, 99 - j, :]
        return A1

    def forward(self, sentence, query_sentence, M_XOR, M_cos, masks):
        """
        M_XOR or M_cos: (B, Q, D)
        masks: (B, Q, D) have 0 on non-pad positions and 1 on pad positions
        """
        sentence, query_sentence = sentence.to(device), query_sentence.to(device)

        x, query_x = self.embedding(sentence), self.embedding(query_sentence)

        X_i = self.Ws(query_x).view(self.batch_size, -1)
        Y_j = self.Ws(x).view(self.batch_size, -1)

        total_passage_level = torch.randn(self.passagelen, self.batch_size, 8).type(torch.FloatTensor).to(device)
        M_cos_passage = torch.randn(self.passagelen, self.batch_size, self.p["maxqlen"], 100).type(torch.FloatTensor).to(device)
        M_XOR_passage = torch.randn(self.passagelen, self.batch_size, self.p["maxqlen"], 100).type(torch.FloatTensor).to(device)
        Y_j_passage = torch.randn(self.passagelen, self.batch_size, 100).type(torch.FloatTensor).to(device)
        X_i_passage = torch.randn(self.passagelen, self.batch_size, self.p["maxqlen"]).type(torch.FloatTensor).to(device)

        mask_passage = torch.randn(self.passagelen, self.batch_size, self.p["maxqlen"], 100).type(torch.FloatTensor).to(device)

        for number_window in range(self.passagelen):
            mask_passage[number_window] = masks[:, :, (number_window * 100) : ((number_window + 1) * 100)]  # (P, BAT, Q, 100)

            M_cos_passage[number_window] = M_cos[:, :, (number_window * 100) : ((number_window + 1) * 100)]  # (P, BAT, Q, 100)
            M_XOR_passage[number_window] = M_XOR[:, :, (number_window * 100) : ((number_window + 1) * 100)]  # (P, BAT, Q, 100)
            Y_j_passage[number_window] = Y_j[:, (number_window * 100) : ((number_window + 1) * 100)]  # (P, BAT, 100)
            X_i_passage[number_window] = X_i  # (P, BAT, 100)

        S_cos = (
            torch.randn(self.passagelen, self.batch_size, self.p["maxqlen"], 100, 3).type(torch.FloatTensor).to(device)
        )  # (P, BAT, Q, 100, 3)
        S_xor = torch.randn(self.passagelen, self.batch_size, self.p["maxqlen"], 100, 3).type(torch.FloatTensor).to(device)
        S_cos[:, :, :, :, 0] = X_i_passage.reshape(self.passagelen, self.batch_size, self.p["maxqlen"], 1).expand(
            self.passagelen, self.batch_size, self.p["maxqlen"], 100
        )
        S_cos[:, :, :, :, 1] = Y_j_passage.reshape(self.passagelen, self.batch_size, 1, 100).expand(
            self.passagelen, self.batch_size, self.p["maxqlen"], 100
        )
        S_cos[:, :, :, :, 2] = M_cos_passage
        S_xor[:, :, :, :, 0] = X_i_passage.reshape(self.passagelen, self.batch_size, self.p["maxqlen"], 1).expand(
            self.passagelen, self.batch_size, self.p["maxqlen"], 100
        )
        S_xor[:, :, :, :, 1] = Y_j_passage.reshape(self.passagelen, self.batch_size, 1, 100).expand(
            self.passagelen, self.batch_size, self.p["maxqlen"], 100
        )
        S_xor[:, :, :, :, 2] = M_XOR_passage

        # add mask on X and Y
        S_cos = S_cos * (1 - mask_passage).unsqueeze(-1)
        S_xor = S_xor * (1 - mask_passage).unsqueeze(-1)

        # S_xor, S_cos: (P, B, Q, 100, 3) -> (P*B, Q, 100, 3)
        S_xor1 = S_xor.view(self.passagelen * self.batch_size, self.p["maxqlen"], 100, 3)
        S_cos1 = S_cos.view(self.passagelen * self.batch_size, self.p["maxqlen"], 100, 3)
        S_xor1_cos1 = torch.cat([S_xor1, S_cos1], dim=0)  # (2*P*B, Q, 100, 3)
        H_xor_cos = self.GRU2d1(S_xor1_cos1)  # (2*P*B, 2)

        H_xor = H_xor_cos[: (self.passagelen * self.batch_size)]  # (P*B, 2)
        H_cos = H_xor_cos[(self.passagelen * self.batch_size) :]  # (P*B, 2)
        e = torch.cat([H_xor, H_cos], dim=-1)  # (P*B, 4)

        S_xor_inv = self.matrix_inv(S_xor1)  # (P*B, Q, 100, 3)
        S_cos_inv = self.matrix_inv(S_cos1)
        S_xor_cos_inv = torch.cat([S_xor_inv, S_cos_inv], dim=0)
        H_xor_cos_inv = self.GRU2d3(S_xor_cos_inv)
        H_xor_inv = H_xor_cos_inv[: (self.passagelen * self.batch_size)]
        H_cos_inv = H_xor_cos_inv[(self.passagelen * self.batch_size) :]

        e_inv = torch.cat([H_xor_inv, H_cos_inv], dim=-1)  # (P*B, 4)
        passage_level_e = torch.cat([e, e_inv], dim=-1)  # (P*B, 8)
        for number_window in range(self.passagelen):  # (P, B, 8)
            total_passage_level[number_window] = passage_level_e[
                (self.batch_size * number_window) : (self.batch_size * (number_window + 1))
            ]

        lstm_out, self.hidden = self.lstm(total_passage_level, self.hidden)
        # lstm_out: (P, B, 2 * self.hidden), where P is the timestep dimension

        lstm_out_forward = lstm_out[:, :, :6]  # (P, B, self.hidden)
        lstm_out_backward = lstm_out[:, :, 6:]  # (P, B, self.hidden)

        # added
        lstm_out = lstm_out_forward + lstm_out_backward

        vt = torch.tanh(self.Wv(total_passage_level))  # (P, B, 8) -> (P, B, 6)

        # evidence = torch.cat((vt, lstm_out_forward, lstm_out_backward), 0)    # (3P, B, 6)
        evidence = torch.cat((vt, lstm_out), 0)  # (2P, B, 6)

        evidence1 = torch.transpose(evidence, 0, 2)  # (6, B, 3P) / (6, B, 2P)
        anss, _ = torch.topk(evidence1, self.p["kmax"], largest=True, sorted=True, dim=2)  # (6, B, 10)
        ans = torch.transpose(anss, 0, 1)  # (B, 6, 10)
        ans1 = ans.contiguous().view(self.batch_size, -1)  # (B, 60)
        score = self.fc(ans1)
        return score.view(-1)


class HiNT_main(nn.Module):
    def __init__(self, extractor, config):
        super(HiNT_main, self).__init__()
        self.HiNT1 = HiNT(extractor.embeddings, config).to(device)
        self.batch_size = config["batch"]

    def init_hidden(self):
        return self.HiNT1.init_hidden()

    def forward(self, query_sentence, query_idf, pos_sentence, neg_sentence):
        self.HiNT1.hidden = self.HiNT1.init_hidden()
        query_sentence = query_sentence
        query_idf = query_idf
        pos_sentence = pos_sentence
        neg_sentence = neg_sentence

        x = self.HiNT1.embedding(pos_sentence)
        query_x = self.HiNT1.embedding(query_sentence)
        BAT = query_sentence.shape[0]
        A = query_sentence.shape[1]
        B = pos_sentence.shape[1]

        x1 = x.norm(dim=2)[:, :, None] + 1e-7
        query_x1 = query_x.norm(dim=2)[:, :, None] + 1e-7
        x_norm = x / x1  # (BAT, B, H)
        query_x_norm = query_x / query_x1  # (BAT, A, H)

        M_cos_pos = torch.matmul(query_x_norm, torch.transpose(x_norm, 1, 2))  # (BAT, A, B)

        nul = torch.zeros_like(M_cos_pos)
        one = torch.ones_like(M_cos_pos)
        XOR_matrix_pos = torch.where(
            query_sentence.reshape(BAT, A, 1).expand(BAT, A, B) == pos_sentence.reshape(BAT, 1, B).expand(BAT, A, B), one, nul
        )

        # add padding for both matrix
        query_masks, sentence_masks = (query_sentence == 0), (pos_sentence == 0)  # (B, Q), (B, D)
        pos_masks = query_masks[:, :, None] * sentence_masks[:, None, :]  # (B, Q, D)
        # add mask
        # XOR_matrix_pos = torch.where(pos_masks, nul, XOR_matrix_pos)
        # M_cos_pos = torch.where(pos_masks, nul, M_cos_pos)
        # mask would be applied in HiNT1
        pos_scores = self.HiNT1(pos_sentence, query_sentence, XOR_matrix_pos, M_cos_pos, pos_masks)

        self.HiNT1.hidden = self.HiNT1.init_hidden()
        x = self.HiNT1.embedding(neg_sentence)
        query_x = self.HiNT1.embedding(query_sentence)
        BAT = query_sentence.shape[0]
        A = query_sentence.shape[1]
        B = neg_sentence.shape[1]

        x1 = x.norm(dim=2)[:, :, None] + 1e-7
        query_x1 = query_x.norm(dim=2)[:, :, None] + 1e-7
        x_norm = x / x1
        query_x_norm = query_x / query_x1
        M_cos_neg = torch.matmul(query_x_norm, torch.transpose(x_norm, 1, 2))
        nul = torch.zeros_like(M_cos_pos)
        one = torch.ones_like(M_cos_pos)
        XOR_matrix_neg = torch.where(
            query_sentence.reshape(BAT, A, 1).expand(BAT, A, B) == neg_sentence.reshape(BAT, 1, B).expand(BAT, A, B), one, nul
        )

        # add mask for both matrix
        query_masks, sentence_masks = (query_sentence == 0), (neg_sentence == 0)  # (B, Q), (B, D)
        neg_masks = query_masks[:, :, None] * sentence_masks[:, None, :]  # (B, Q, D)
        # add mask
        # XOR_matrix_neg = torch.where(neg_masks, nul, XOR_matrix_neg)
        # M_cos_neg = torch.where(neg_masks, nul, M_cos_neg)

        # mask would be applied in HiNT1
        neg_scores = self.HiNT1(neg_sentence, query_sentence, XOR_matrix_neg, M_cos_neg, neg_masks)

        return [pos_scores, neg_scores]

    def test_forward(self, query_sentence, query_idf, pos_sentence):
        self.HiNT1.hidden = self.HiNT1.init_hidden()
        query_sentence = query_sentence
        query_idf = query_idf
        pos_sentence = pos_sentence

        x = self.HiNT1.embedding(pos_sentence)
        query_x = self.HiNT1.embedding(query_sentence)
        BAT = query_sentence.shape[0]
        A = query_sentence.shape[1]
        B = pos_sentence.shape[1]
        x1 = x.norm(dim=2)[:, :, None] + 1e-7
        query_x1 = query_x.norm(dim=2)[:, :, None] + 1e-7
        x_norm = x / x1
        query_x_norm = query_x / query_x1
        M_cos_pos = torch.matmul(query_x_norm, torch.transpose(x_norm, 1, 2))

        nul = torch.zeros_like(M_cos_pos)
        one = torch.ones_like(M_cos_pos)
        XOR_matrix_pos = torch.where(
            query_sentence.reshape(BAT, A, 1).expand(BAT, A, B) == pos_sentence.reshape(BAT, 1, B).expand(BAT, A, B), one, nul
        )

        query_masks, sentence_masks = (query_sentence == 0), (pos_sentence == 0)  # (B, Q), (B, D)
        pos_masks = query_masks[:, :, None] * sentence_masks[:, None, :]  # (B, Q, D)
        # add mask
        # XOR_matrix_pos = torch.where(pos_masks, nul, XOR_matrix_pos)
        # M_cos_pos = torch.where(pos_masks, nul, M_cos_pos)

        # mask would be applied in HiNT1
        pos_scores = self.HiNT1(pos_sentence, query_sentence, XOR_matrix_pos, M_cos_pos, pos_masks)
        return pos_scores


@Reranker.register
class HINT(Reranker):
    """Yixing Fan, Jiafeng Guo, Yanyan Lan, Jun Xu, Chengxiang Zhai, and Xueqi Cheng. 2018. Modeling Diverse Relevance Patterns in Ad-hoc Retrieval. In SIGIR'18."""

    module_name = "HINT"

    dependencies = [
        Dependency(key="extractor", module="extractor", name="slowembedtext"),
        Dependency(key="trainer", module="trainer", name="pytorch"),
    ]
    config_spec = [ConfigOption("spatialGRU", 2), ConfigOption("LSTMdim", 6), ConfigOption("kmax", 10)]

    def test(self, query_sentence, query_idf, pos_sentence, *args, **kwargs):
        return self.model.test_forward(query_sentence, query_idf, pos_sentence)

    def score(self, d):
        query_idf = d["query_idf"]
        query_sentence = d["query"]
        pos_sentence, neg_sentence = d["posdoc"], d["negdoc"]
        return self.model(query_sentence, query_idf, pos_sentence, neg_sentence)

    def test(self, d):
        query_idf = d["query_idf"]
        query_sentence = d["query"]
        pos_sentence = d["posdoc"]

        return self.model.test_forward(query_sentence, query_idf, pos_sentence)

    def zero_grad(self, *args, **kwargs):
        self.model.zero_grad(*args, **kwargs)

    def build_model(self):
        if not hasattr(self, "model"):
            config = dict(self.config)
            config.update(self.extractor.config)
            config["batch"] = self.trainer.config["batch"]
            self.model = HiNT_main(self.extractor, config)

        return self.model



import tensorflow as tf
from transformers import TFBertForSequenceClassification

from capreolus import ConfigOption, Dependency
from capreolus.reranker import Reranker
from capreolus.utils.loginit import get_logger

logger = get_logger(__name__)


class TFVanillaBert_Class(tf.keras.layers.Layer):
    def __init__(self, extractor, config, *args, **kwargs):
        super(TFVanillaBert_Class, self).__init__(*args, **kwargs)
        self.extractor = extractor

        # TFBertForSequenceClassification contains both the BERT and the linear classifier layers
        self.bert = TFBertForSequenceClassification.from_pretrained(config["pretrained"], hidden_dropout_prob=0.1)

        assert extractor.config["numpassages"] == 1, "numpassages should be 1 for TFVanillaBERT"
        self.config = config

    def call(self, x, **kwargs):
        """
        Returns logits of shape [2]
        """
        doc_bert_input, doc_mask, doc_seg = x[0], x[1], x[2]
        doc_scores = self.bert(doc_bert_input, attention_mask=doc_mask, token_type_ids=doc_seg)[0]

        return doc_scores

    def predict_step(self, data):
        posdoc_bert_input, posdoc_mask, posdoc_seg, negdoc_bert_input, negdoc_mask, negdoc_seg = data
        batch_size = tf.shape(posdoc_bert_input)[0]
        num_passages = tf.shape(posdoc_bert_input)[1]
        tf.debugging.assert_equal(num_passages, 1)
        maxseqlen = self.extractor.config["maxseqlen"]

        posdoc_bert_input = tf.reshape(posdoc_bert_input, [batch_size * num_passages, maxseqlen])
        posdoc_mask = tf.reshape(posdoc_mask, [batch_size * num_passages, maxseqlen])
        posdoc_seg = tf.reshape(posdoc_seg, [batch_size * num_passages, maxseqlen])

        doc_scores = self.call((posdoc_bert_input, posdoc_mask, posdoc_seg), training=False)[:, 1]
        tf.debugging.assert_equal(tf.shape(doc_scores), [batch_size])

        return doc_scores

    def score(self, x, **kwargs):
        posdoc_bert_input, posdoc_mask, posdoc_seg, negdoc_bert_input, negdoc_mask, negdoc_seg = x

        return self.call((posdoc_bert_input, posdoc_mask, posdoc_seg), **kwargs)

    def score_pair(self, x, **kwargs):
        posdoc_bert_input, posdoc_mask, posdoc_seg, negdoc_bert_input, negdoc_mask, negdoc_seg = x

        pos_score = self.call((posdoc_bert_input, posdoc_mask, posdoc_seg), **kwargs)[:, 1]
        neg_score = self.call((negdoc_bert_input, negdoc_mask, negdoc_seg), **kwargs)[:, 1]

        return pos_score, neg_score


@Reranker.register
class TFVanillaBERT(Reranker):
    """
    TensorFlow implementation of Vanilla BERT.
    Input is of the form [CLS] sentence A [SEP] sentence B [SEP]
    The "score" of a query (sentence A) - document (sentence B) pair is the probability that the document is relevant
    to the query. This is achieved through a linear classifier layer attached to BERT's last layer and using the logits[1] as the score.
    """

    module_name = "TFVanillaBERT"

    dependencies = [
        Dependency(key="extractor", module="extractor", name="bertpassage"),
        Dependency(key="trainer", module="trainer", name="tensorflow"),
    ]
    config_spec = [ConfigOption("pretrained", "bert-base-uncased", "pretrained model to load")]

    def build_model(self):
        self.model = TFVanillaBert_Class(self.extractor, self.config)
        return self.model



import json
import re

from capreolus import ConfigOption, Dependency, constants
from capreolus.utils.loginit import get_logger
from capreolus.utils.trec import topic_to_trectxt

from . import Benchmark

logger = get_logger(__name__)
PACKAGE_PATH = constants["PACKAGE_PATH"]


@Benchmark.register
class NF(Benchmark):
    """NFCorpus: A Full-Text Learning to Rank Dataset for Medical Information Retrieval [1]

    [1] Vera Boteva, Demian Gholipour, Artem Sokolov and Stefan Riezler. A Full-Text Learning to Rank Dataset for Medical Information Retrieval Proceedings of the 38th European Conference on Information Retrieval (ECIR), Padova, Italy, 2016. https://www.cl.uni-heidelberg.de/statnlpgroup/nfcorpus/
    """

    module_name = "nf"
    dependencies = [Dependency(key="collection", module="collection", name="nf")]
    config_spec = [
        ConfigOption(key="labelrange", default_value="0-2", description="range of dataset qrels, options: 0-2, 1-3"),
        ConfigOption(
            key="fields",
            default_value="all_titles",
            description="query fields included in topic file, "
            "options: 'all_fields', 'all_titles', 'nontopics', 'vid_title', 'vid_desc'",
        ),
    ]

    fold_file = PACKAGE_PATH / "data" / "nf.json"

    query_type = "title"

    def build(self):
        fields, label_range = self.config["fields"], self.config["labelrange"]
        self.field2kws = {
            "all_fields": ["all"],
            "nontopics": ["nontopic-titles"],
            "vid_title": ["vid-titles"],
            "vid_desc": ["vid-desc"],
            "all_titles": ["titles", "vid-titles", "nontopic-titles"],
        }
        self.labelrange2kw = {"0-2": "2-1-0", "1-3": "3-2-1"}

        if fields not in self.field2kws:
            raise ValueError(f"Unexpected fields value: {fields}, expect: {', '.join(self.field2kws.keys())}")
        if label_range not in self.labelrange2kw:
            raise ValueError(f"Unexpected label range: {label_range}, expect: {', '.join(self.field2kws.keys())}")

        self.qrel_file = PACKAGE_PATH / "data" / f"qrels.nf.{label_range}.txt"
        self.test_qrel_file = PACKAGE_PATH / "data" / f"test.qrels.nf.{label_range}.txt"
        self.topic_file = PACKAGE_PATH / "data" / f"topics.nf.{fields}.txt"
        self.download_if_missing()

    def _transform_qid(self, raw):
        """ NFCorpus dataset specific, remove prefix in query id since anserini convert all qid to integer """
        return raw.replace("PLAIN-", "")

    def download_if_missing(self):
        if all([f.exists() for f in [self.topic_file, self.fold_file, self.qrel_file]]):
            return

        tmp_corpus_dir = self.collection.download_raw()
        topic_f = open(self.topic_file, "w", encoding="utf-8")
        qrel_f = open(self.qrel_file, "w", encoding="utf-8")
        test_qrel_f = open(self.test_qrel_file, "w", encoding="utf-8")

        set_names = ["train", "dev", "test"]
        folds = {s: set() for s in set_names}
        qrel_kw = self.labelrange2kw[self.config["labelrange"]]
        for set_name in set_names:
            with open(tmp_corpus_dir / f"{set_name}.{qrel_kw}.qrel") as f:
                for line in f:
                    line = self._transform_qid(line)
                    qid = line.strip().split()[0]
                    folds[set_name].add(qid)
                    if set_name == "test":
                        test_qrel_f.write(line)
                    qrel_f.write(line)

            files = [tmp_corpus_dir / f"{set_name}.{keyword}.queries" for keyword in self.field2kws[self.config["fields"]]]
            qids2topics = self._align_queries(files, "title")

            for qid, txts in qids2topics.items():
                topic_f.write(topic_to_trectxt(qid, txts["title"]))

        json.dump(
            {"s1": {"train_qids": list(folds["train"]), "predict": {"dev": list(folds["dev"]), "test": list(folds["test"])}}},
            open(self.fold_file, "w"),
        )

        topic_f.close()
        qrel_f.close()
        test_qrel_f.close()
        logger.info(f"nf benchmark prepared")

    def _align_queries(self, files, field, qid2queries=None):
        if not qid2queries:
            qid2queries = {}
        for fn in files:
            with open(fn, "r", encoding="utf-8") as f:
                for line in f:
                    qid, txt = line.strip().split("\t")
                    qid = self._transform_qid(qid)
                    txt = " ".join(re.sub("[^A-Za-z]", " ", txt).split()[:1020])
                    if qid not in qid2queries:
                        qid2queries[qid] = {field: txt}
                    else:
                        if field in qid2queries[qid]:
                            logger.warning(f"Overwriting title for query {qid}")
                        qid2queries[qid][field] = txt
        return qid2queries



import os
import shutil

from capreolus import constants
from capreolus.utils.common import download_file, hash_file
from capreolus.utils.loginit import get_logger

from . import Collection

logger = get_logger(__name__)
PACKAGE_PATH = constants["PACKAGE_PATH"]


@Collection.register
class ANTIQUE(Collection):
    """A Non-factoid Question Answering Benchmark from Hashemi et al. [1]

    [1] Helia Hashemi, Mohammad Aliannejadi, Hamed Zamani, and W. Bruce Croft. 2020. ANTIQUE: A non-factoid question answering benchmark. ECIR 2020.
    """

    module_name = "antique"
    _path = PACKAGE_PATH / "data" / "antique-collection"

    collection_type = "TrecCollection"
    generator_type = "DefaultLuceneDocumentGenerator"

    def download_if_missing(self):
        url = "http://ciir.cs.umass.edu/downloads/Antique/antique-collection.txt"
        cachedir = self.get_cache_path()
        document_dir = os.path.join(cachedir, "documents")
        coll_filename = os.path.join(document_dir, "antique-collection.txt")

        if os.path.exists(coll_filename):
            return document_dir

        tmp_dir = cachedir / "tmp"
        tmp_filename = os.path.join(tmp_dir, "tmp.anqique.file")

        os.makedirs(tmp_dir, exist_ok=True)
        os.makedirs(document_dir, exist_ok=True)

        download_file(url, tmp_filename, expected_hash="68b6688f5f2668c93f0e8e43384f66def768c4da46da4e9f7e2629c1c47a0c36")
        self._convert_to_trec(inp_path=tmp_filename, outp_path=coll_filename)
        logger.info(f"antique collection file prepared, stored at {coll_filename}")

        for file in os.listdir(tmp_dir):  # in case there are legacy files
            os.remove(os.path.join(tmp_dir, file))
        shutil.rmtree(tmp_dir)

        return document_dir

    def _convert_to_trec(self, inp_path, outp_path):
        assert os.path.exists(inp_path)

        fout = open(outp_path, "wt", encoding="utf-8")
        with open(inp_path, "rt", encoding="utf-8") as f:
            for line in f:
                docid, doc = line.strip().split("\t")
                fout.write(f"<DOC>\n<DOCNO>{docid}</DOCNO>\n<TEXT>\n{doc}\n</TEXT>\n</DOC>\n")
        fout.close()
        logger.debug(f"Converted file {os.path.basename(inp_path)} to TREC format, output to: {outp_path}")

    def _validate_document_path(self, path):
        """ Checks that the sha256sum is correct """
        return (
            hash_file(os.path.join(path, "antique-collection.txt"))
            == "409e0960f918970977ceab9e5b1d372f45395af25d53b95644bdc9ccbbf973da"
        )



from capreolus import ConfigOption

from . import Tokenizer


@Tokenizer.register
class AnseriniTokenizer(Tokenizer):
    module_name = "anserini"
    config_spec = [
        ConfigOption("keepstops", True, "keep stopwords if True"),
        ConfigOption("stemmer", "none", "stemmer: porter, krovetz, or none"),
    ]

    def build(self):
        self._tokenize = self._get_tokenize_fn()

    def _get_tokenize_fn(self):
        from jnius import autoclass

        stemmer, keepstops = self.config["stemmer"], self.config["keepstops"]
        if stemmer is None:
            stemmer = "none"

        emptyjchar = autoclass("org.apache.lucene.analysis.CharArraySet").EMPTY_SET
        Analyzer = autoclass("io.anserini.analysis.DefaultEnglishAnalyzer")
        analyzer = Analyzer.newStemmingInstance(stemmer, emptyjchar) if keepstops else Analyzer.newStemmingInstance(stemmer)
        tokenizefn = autoclass("io.anserini.analysis.AnalyzerUtils").analyze

        def _tokenize(sentence):
            return tokenizefn(analyzer, sentence).toArray()

        return _tokenize

    def tokenize(self, sentences):
        if not sentences or len(sentences) == 0:  # either "" or []
            return []

        if isinstance(sentences, str):
            return self._tokenize(sentences)

        return [self._tokenize(s) for s in sentences]



import tensorflow as tf
import numpy as np

from capreolus import Dependency, ConfigOption, get_logger
from capreolus.utils.common import padlist
from capreolus.utils.exceptions import MissingDocError
from . import Extractor
from .bertpassage import BertPassage

logger = get_logger(__name__)


@Extractor.register
class PooledBertPassage(BertPassage):
    """
    Extracts passages from the document to be later consumed by a BERT based model.
    Different from BertPassage in the sense that all the passages from a document "stick together" during training - the
    resulting feature always have the shape (batch, num_passages, maxseqlen) - and this allows the reranker to pool
    over passages from the same document during training

    """

    module_name = "pooledbertpassage"
    dependencies = [
        Dependency(
            key="index", module="index", name="anserini", default_config_overrides={"indexstops": True, "stemmer": "none"}
        ),
        Dependency(key="tokenizer", module="tokenizer", name="berttokenizer"),
    ]

    pad = 0
    pad_tok = "[PAD]"

    config_spec = [
        ConfigOption("maxseqlen", 256, "Maximum input length for BERT"),
        ConfigOption("usecache", False, "Should the extracted features be cached?"),
        ConfigOption("passagelen", 150, "Length of the extracted passage"),
        ConfigOption("stride", 100, "Stride"),
        ConfigOption("sentences", False, "Use a sentence tokenizer to form passages"),
        ConfigOption("numpassages", 16, "Number of passages per document"),
        ConfigOption(
            "prob",
            0.1,
            "The probability that a passage from the document will be used for training (the first passage is always used)",
        ),
    ]

    def create_tf_train_feature(self, sample):
        """
        Returns a set of features from a doc.
        Of the num_passages passages that are present in a document, we use only a subset of it.
        params:
        sample - A dict where each entry has the shape [batch_size, num_passages, maxseqlen]

        Returns a list of features. Each feature is a dict, and each value in the dict has the shape [batch_size, maxseqlen].
        Yes, the output shape is different to the input shape because we sample from the passages.
        """
        return self.create_tf_dev_feature(sample)

    def create_tf_dev_feature(self, sample):
        """
        Unlike the train feature, the dev set uses all passages. Both the input and the output are dicts with the shape
        [batch_size, num_passages, maxseqlen]
        """
        posdoc, negdoc, negdoc_id = sample["pos_bert_input"], sample["neg_bert_input"], sample["negdocid"]
        posdoc_mask, posdoc_seg, negdoc_mask, negdoc_seg = (
            sample["pos_mask"],
            sample["pos_seg"],
            sample["neg_mask"],
            sample["neg_seg"],
        )
        label = sample["label"]

        def _bytes_feature(value):
            """Returns a bytes_list from a string / byte."""
            if isinstance(value, type(tf.constant(0))):  # if value ist tensor
                value = value.numpy()  # get value of tensor
            return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))

        feature = {
            "pos_bert_input": _bytes_feature(tf.io.serialize_tensor(posdoc)),
            "pos_mask": _bytes_feature(tf.io.serialize_tensor(posdoc_mask)),
            "pos_seg": _bytes_feature(tf.io.serialize_tensor(posdoc_seg)),
            "neg_bert_input": _bytes_feature(tf.io.serialize_tensor(negdoc)),
            "neg_mask": _bytes_feature(tf.io.serialize_tensor(negdoc_mask)),
            "neg_seg": _bytes_feature(tf.io.serialize_tensor(negdoc_seg)),
            "label": _bytes_feature(tf.io.serialize_tensor(label)),
        }

        return [feature]

    def parse_tf_train_example(self, example_proto):
        return self.parse_tf_dev_example(example_proto)

    def parse_tf_dev_example(self, example_proto):
        feature_description = self.get_tf_feature_description()
        parsed_example = tf.io.parse_example(example_proto, feature_description)

        def parse_tensor_as_int(x):
            parsed_tensor = tf.io.parse_tensor(x, tf.int64)
            parsed_tensor.set_shape([self.config["numpassages"], self.config["maxseqlen"]])

            return parsed_tensor

        def parse_label_tensor(x):
            parsed_tensor = tf.io.parse_tensor(x, tf.float32)
            parsed_tensor.set_shape([2])

            return parsed_tensor

        pos_bert_input = tf.map_fn(parse_tensor_as_int, parsed_example["pos_bert_input"], dtype=tf.int64)
        pos_mask = tf.map_fn(parse_tensor_as_int, parsed_example["pos_mask"], dtype=tf.int64)
        pos_seg = tf.map_fn(parse_tensor_as_int, parsed_example["pos_seg"], dtype=tf.int64)
        neg_bert_input = tf.map_fn(parse_tensor_as_int, parsed_example["neg_bert_input"], dtype=tf.int64)
        neg_mask = tf.map_fn(parse_tensor_as_int, parsed_example["neg_mask"], dtype=tf.int64)
        neg_seg = tf.map_fn(parse_tensor_as_int, parsed_example["neg_seg"], dtype=tf.int64)
        label = tf.map_fn(parse_label_tensor, parsed_example["label"], dtype=tf.float32)

        return (pos_bert_input, pos_mask, pos_seg, neg_bert_input, neg_mask, neg_seg), label

    def id2vec(self, qid, posid, negid=None, label=None):
        """
        See parent class for docstring
        """
        assert label is not None

        tokenizer = self.tokenizer
        maxseqlen = self.config["maxseqlen"]

        query_toks = self.qid2toks[qid]
        pos_bert_inputs = []
        pos_bert_masks = []
        pos_bert_segs = []

        # N.B: The passages in self.docid2passages are not bert tokenized
        pos_passages = self.docid2passages[posid]
        for tokenized_passage in pos_passages:
            input_line = ["[CLS]"] + query_toks + ["[SEP]"] + tokenized_passage + ["[SEP]"]
            if len(input_line) > maxseqlen:
                input_line = input_line[:maxseqlen]
                input_line[-1] = "[SEP]"

            padded_input_line = padlist(input_line, padlen=self.config["maxseqlen"], pad_token=self.pad_tok)
            pos_bert_masks.append([1] * len(input_line) + [0] * (len(padded_input_line) - len(input_line)))
            pos_bert_segs.append([0] * (len(query_toks) + 2) + [1] * (len(padded_input_line) - len(query_toks) - 2))
            pos_bert_inputs.append(tokenizer.convert_tokens_to_ids(padded_input_line))

        # TODO: Rename the posdoc key in the below dict to 'pos_bert_input'
        data = {
            "qid": qid,
            "posdocid": posid,
            "pos_bert_input": np.array(pos_bert_inputs, dtype=np.long),
            "pos_mask": np.array(pos_bert_masks, dtype=np.long),
            "pos_seg": np.array(pos_bert_segs, dtype=np.long),
            "negdocid": "",
            "neg_bert_input": np.zeros((self.config["numpassages"], self.config["maxseqlen"]), dtype=np.long),
            "neg_mask": np.zeros((self.config["numpassages"], self.config["maxseqlen"]), dtype=np.long),
            "neg_seg": np.zeros((self.config["numpassages"], self.config["maxseqlen"]), dtype=np.long),
            "label": np.array(label, dtype=np.float32),
        }

        if negid:
            neg_bert_inputs = []
            neg_bert_masks = []
            neg_bert_segs = []
            neg_passages = self.docid2passages[negid]
            for tokenized_passage in neg_passages:
                input_line = ["[CLS]"] + query_toks + ["[SEP]"] + tokenized_passage + ["[SEP]"]
                if len(input_line) > maxseqlen:
                    input_line = input_line[:maxseqlen]
                    input_line[-1] = "[SEP]"

                padded_input_line = padlist(input_line, padlen=self.config["maxseqlen"], pad_token=self.pad_tok)
                neg_bert_masks.append([1] * len(input_line) + [0] * (len(padded_input_line) - len(input_line)))
                neg_bert_segs.append([0] * (len(query_toks) + 2) + [1] * (len(padded_input_line) - len(query_toks) - 2))
                neg_bert_inputs.append(tokenizer.convert_tokens_to_ids(padded_input_line))

            if not neg_bert_inputs:
                raise MissingDocError(qid, negid)

            data["negdocid"] = negid
            data["neg_bert_input"] = np.array(neg_bert_inputs, dtype=np.long)
            data["neg_mask"] = np.array(neg_bert_masks, dtype=np.long)
            data["neg_seg"] = np.array(neg_bert_segs, dtype=np.long)

        return data



import os
import tarfile

from capreolus import constants, get_logger
from capreolus.utils.common import download_file

from . import Collection

logger = get_logger(__name__)
PACKAGE_PATH = constants["PACKAGE_PATH"]


@Collection.register
class NF(Collection):
    """NFCorpus: A Full-Text Learning to Rank Dataset for Medical Information Retrieval [1]

    [1] Vera Boteva, Demian Gholipour, Artem Sokolov and Stefan Riezler. A Full-Text Learning to Rank Dataset for Medical Information Retrieval Proceedings of the 38th European Conference on Information Retrieval (ECIR), Padova, Italy, 2016. https://www.cl.uni-heidelberg.de/statnlpgroup/nfcorpus/
    """

    module_name = "nf"
    _path = PACKAGE_PATH / "data" / "nf-collection"
    url = "http://www.cl.uni-heidelberg.de/statnlpgroup/nfcorpus/nfcorpus.tar.gz"

    collection_type = "TrecCollection"
    generator_type = "DefaultLuceneDocumentGenerator"

    def download_raw(self):
        cachedir = self.get_cache_path()
        tmp_dir = cachedir / "tmp"
        tmp_tar_fn, tmp_corpus_dir = tmp_dir / "nfcorpus.tar.gz", tmp_dir / "nfcorpus"

        os.makedirs(tmp_dir, exist_ok=True)

        if not tmp_tar_fn.exists():
            download_file(self.url, tmp_tar_fn, "ebc026d4a8bef3f866148b727e945a2073eb4045ede9b7de95dd50fd086b4256")

        with tarfile.open(tmp_tar_fn) as f:
            f.extractall(tmp_dir)
        return tmp_corpus_dir

    def download_if_missing(self):
        cachedir = self.get_cache_path()
        document_dir = os.path.join(cachedir, "documents")
        coll_filename = os.path.join(document_dir, "nf-collection.txt")
        if os.path.exists(coll_filename):
            return document_dir

        os.makedirs(document_dir, exist_ok=True)
        tmp_corpus_dir = self.download_raw()

        inp_fns = [tmp_corpus_dir / f"{set_name}.docs" for set_name in ["train", "dev", "test"]]
        print(inp_fns)
        with open(coll_filename, "w", encoding="utf-8") as outp_file:
            self._convert_to_trec(inp_fns, outp_file)
        logger.info(f"nf collection file prepared, stored at {coll_filename}")

        return document_dir

    def _convert_to_trec(self, inp_fns, outp_file):
        # train.docs, dev.docs, and test.docs have some overlap, so we check for duplicate docids
        seen_docids = set()

        for inp_fn in inp_fns:
            assert os.path.exists(inp_fn)

            with open(inp_fn, "rt", encoding="utf-8") as f:
                for line in f:
                    docid, doc = line.strip().split("\t")

                    if docid not in seen_docids:
                        outp_file.write(f"<DOC>\n<DOCNO>{docid}</DOCNO>\n<TEXT>\n{doc}\n</TEXT>\n</DOC>\n")
                        seen_docids.add(docid)



import gzip
import os
import xml.etree.ElementTree as ET
from collections import defaultdict


def load_ntcir_topics(fn):
    topics = {}

    tree = ET.parse(fn)
    for child in tree.getroot():
        qid = child.find("qid").text.strip()
        query = child.find("content").text.strip()

        assert qid not in topics
        assert len(qid) > 0 and len(query) > 0
        topics[qid] = query

    return {"content": topics}


def load_trec_topics(queryfn):
    title, desc, narr = defaultdict(list), defaultdict(list), defaultdict(list)

    block = None
    if str(queryfn).endswith(".gz"):
        openf = gzip.open
    else:
        openf = open

    with openf(queryfn, "rt") as f:
        for line in f:
            line = line.strip()

            if line.startswith("<num>"):
                # <num> Number: 700
                qid = line.split()[-1]
                # no longer an int
                # assert qid > 0
                block = None
            elif line.startswith("<title>"):
                # <title>  query here
                title[qid].extend(line.strip().split()[1:])
                block = "title"
                # TODO does this sometimes start with Topic: ?
                assert "Topic:" not in line
            elif line.startswith("<desc>"):
                # <desc> description \n description
                desc[qid].extend(line.strip().split()[1:])
                block = "desc"
            elif line.startswith("<narr>"):
                # same format as <desc>
                narr[qid].extend(line.strip().split()[1:])
                block = "narr"
            elif line.startswith("</top>") or line.startswith("<top>"):
                block = None
            elif block == "title":
                title[qid].extend(line.strip().split())
            elif block == "desc":
                desc[qid].extend(line.strip().split())
            elif block == "narr":
                narr[qid].extend(line.strip().split())

    out = {}
    if len(title) > 0:
        out["title"] = {qid: " ".join(terms) for qid, terms in title.items()}
    if len(desc) > 0:
        out["desc"] = {qid: " ".join(terms) for qid, terms in desc.items()}
    if len(narr) > 0:
        out["narr"] = {qid: " ".join(terms) for qid, terms in narr.items()}

    return out


def load_qrels(qrelfile, qids=None, include_spam=True):
    labels = defaultdict(dict)
    with open(qrelfile, "rt") as f:
        for line in f:
            line = line.strip()
            if len(line) == 0:
                continue

            cols = line.split()
            qid, docid, label = cols[0], cols[2], int(cols[3])

            if qids is not None and qid not in qids:
                continue
            if label < 0 and not include_spam:
                continue

            labels[qid][docid] = label

    # remove qids with no relevant docs
    for qid in list(labels.keys()):
        if max(labels[qid].values()) <= 0:
            del labels[qid]

    labels.default_factory = None  # behave like normal dict
    return labels


def document_to_trectxt(docno, txt):
    s = f"<DOC>\n<DOCNO> {docno} </DOCNO>\n"
    s += f"<TEXT>\n{txt}\n</TEXT>\n</DOC>\n"
    return s


def topic_to_trectxt(qno, title, desc=None, narr=None):
    return (
        f"<top>\n\n"
        f"<num> Number: {qno}\n"
        f"<title> {title}\n\n"
        f"<desc> Description:\n{desc or title}\n\n"
        f"<narr> Narrative:\n{narr or title}\n\n"
        f"</top>\n\n\n"
    )


def anserini_index_to_trec_docs(index_dir, output_dir, expected_doc_count):
    from jnius import autoclass

    JFile = autoclass("java.io.File")
    JFSDirectory = autoclass("org.apache.lucene.store.FSDirectory")
    JIndexReaderUtils = autoclass("io.anserini.index.IndexReaderUtils")
    JIndexUtils = autoclass("io.anserini.index.IndexUtils")
    index_utils = JIndexUtils(index_dir)

    index_reader_utils = JIndexReaderUtils()

    fsdir = JFSDirectory.open(JFile(index_dir).toPath())
    reader = autoclass("org.apache.lucene.index.DirectoryReader").open(fsdir)

    docids = set()
    for i in range(expected_doc_count):
        try:
            docid = index_reader_utils.convertLuceneDocidToDocid(reader, i)
            docids.add(docid)
        except:  # lgtm [py/catch-base-exception]
            # we reached the end?
            pass

    if len(docids) != expected_doc_count:
        raise ValueError(
            f"we expected to retrieve {expected_doc_count} documents from the index, but actually found {len(docids)}"
        )

    output_handles = [gzip.open(os.path.join(output_dir, f"{i}.gz"), "wt", encoding="utf-8") for i in range(100, 200)]

    for docidx, docid in enumerate(sorted(docids)):
        txt = document_to_trectxt(docid, index_utils.getRawDocument(docid))
        handleidx = docidx % len(output_handles)
        print(txt, file=output_handles[handleidx])

    for handle in output_handles:
        handle.close()



import math
import os
import time

import numpy as np
import torch
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

from capreolus import ConfigOption, Searcher, constants, evaluator, get_logger
from capreolus.reranker.common import pair_hinge_loss, pair_softmax_loss

from . import Trainer

logger = get_logger(__name__)  # pylint: disable=invalid-name
RESULTS_BASE_PATH = constants["RESULTS_BASE_PATH"]


@Trainer.register
class PytorchTrainer(Trainer):
    module_name = "pytorch"
    config_spec = [
        ConfigOption("batch", 32, "batch size"),
        ConfigOption("niters", 20, "number of iterations to train for"),
        ConfigOption("itersize", 512, "number of training instances in one iteration"),
        ConfigOption("gradacc", 1, "number of batches to accumulate over before updating weights"),
        ConfigOption("lr", 0.001, "learning rate"),
        ConfigOption("softmaxloss", False, "True to use softmax loss (over pairs) or False to use hinge loss"),
        ConfigOption("fastforward", False),
        ConfigOption("validatefreq", 1),
        ConfigOption(
            "multithread",
            False,
            "True to load data in a separate thread; faster but causes PyTorch deadlock in some environments",
        ),
        ConfigOption("boardname", "default"),
    ]
    config_keys_not_in_path = ["fastforward", "boardname"]

    def build(self):
        # sanity checks
        if self.config["batch"] < 1:
            raise ValueError("batch must be >= 1")

        if self.config["niters"] <= 0:
            raise ValueError("niters must be > 0")

        if self.config["itersize"] < self.config["batch"]:
            raise ValueError("itersize must be >= batch")

        if self.config["gradacc"] < 1 or not float(self.config["gradacc"]).is_integer():
            raise ValueError("gradacc must be an integer >= 1")

        if self.config["lr"] <= 0:
            raise ValueError("lr must be > 0")

        torch.manual_seed(self.config["seed"])
        torch.cuda.manual_seed_all(self.config["seed"])

    def single_train_iteration(self, reranker, train_dataloader):
        """Train model for one iteration using instances from train_dataloader.

        Args:
           model (Reranker): a PyTorch Reranker
           train_dataloader (DataLoader): a PyTorch DataLoader that iterates over training instances

        Returns:
            float: average loss over the iteration

        """

        iter_loss = []
        batches_since_update = 0
        batches_per_epoch = (self.config["itersize"] // self.config["batch"]) or 1
        batches_per_step = self.config["gradacc"]

        for bi, batch in tqdm(enumerate(train_dataloader), desc="Iter progression", total=batches_per_epoch):
            # TODO make sure _prepare_batch_with_strings equivalent is happening inside the sampler
            batch = {k: v.to(self.device) if not isinstance(v, list) else v for k, v in batch.items()}
            doc_scores = reranker.score(batch)
            loss = self.loss(doc_scores)
            iter_loss.append(loss)
            loss.backward()

            batches_since_update += 1
            if batches_since_update == batches_per_step:
                batches_since_update = 0
                self.optimizer.step()
                self.optimizer.zero_grad()

            if (bi + 1) % batches_per_epoch == 0:
                break

        return torch.stack(iter_loss).mean()

    def load_loss_file(self, fn):
        """Loads loss history from fn

        Args:
           fn (Path): path to a loss.txt file

        Returns:
            a list of losses ordered by iterations

        """

        loss = []
        with fn.open(mode="rt") as f:
            for lineidx, line in enumerate(f):
                line = line.strip()
                if not line:
                    continue

                iteridx, iterloss = line.rstrip().split()

                if int(iteridx) != lineidx:
                    raise IOError(f"malformed loss file {fn} ... did two processes write to it?")

                loss.append(float(iterloss))

        return loss

    def fastforward_training(self, reranker, weights_path, loss_fn):
        """Skip to the last training iteration whose weights were saved.

        If saved model and optimizer weights are available, this method will load those weights into model
        and optimizer, and then return the next iteration to be run. For example, if weights are available for
        iterations 0-10 (11 zero-indexed iterations), the weights from iteration index 10 will be loaded, and
        this method will return 11.

        If an error or inconsistency is encountered when checking for weights, this method returns 0.

        This method checks several files to determine if weights "are available". First, loss_fn is read to
        determine the last recorded iteration. (If a path is missing or loss_fn is malformed, 0 is returned.)
        Second, the weights from the last recorded iteration in loss_fn are loaded into the model and optimizer.
        If this is successful, the method returns `1 + last recorded iteration`. If not, it returns 0.
        (We consider loss_fn because it is written at the end of every training iteration.)

        Args:
           model (Reranker): a PyTorch Reranker whose state should be loaded
           weights_path (Path): directory containing model and optimizer weights
           loss_fn (Path): file containing loss history

        Returns:
            int: the next training iteration after fastforwarding. If successful, this is > 0.
                 If no weights are available or they cannot be loaded, 0 is returned.

        """

        if not (weights_path.exists() and loss_fn.exists()):
            return 0

        try:
            loss = self.load_loss_file(loss_fn)
        except IOError:
            return 0

        last_loss_iteration = len(loss) - 1
        weights_fn = weights_path / f"{last_loss_iteration}.p"

        try:
            reranker.load_weights(weights_fn, self.optimizer)
            return last_loss_iteration + 1
        except:  # lgtm [py/catch-base-exception]
            logger.info("attempted to load weights from %s but failed, starting at iteration 0", weights_fn)
            return 0

    def train(self, reranker, train_dataset, train_output_path, dev_data, dev_output_path, qrels, metric, relevance_level=1):
        """Train a model following the trainer's config (specifying batch size, number of iterations, etc).

        Args:
           train_dataset (IterableDataset): training dataset
           train_output_path (Path): directory under which train_dataset runs and training loss will be saved
           dev_data (IterableDataset): dev dataset
           dev_output_path (Path): directory where dev_data runs and metrics will be saved

        """
        # Set up logging
        # TODO why not put this under train_output_path?
        summary_writer = SummaryWriter(RESULTS_BASE_PATH / "runs" / self.config["boardname"], comment=train_output_path)

        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        model = reranker.model.to(self.device)
        self.optimizer = torch.optim.Adam(filter(lambda param: param.requires_grad, model.parameters()), lr=self.config["lr"])

        if self.config["softmaxloss"]:
            self.loss = pair_softmax_loss
        else:
            self.loss = pair_hinge_loss

        dev_best_weight_fn, weights_output_path, info_output_path, loss_fn = self.get_paths_for_early_stopping(
            train_output_path, dev_output_path
        )

        initial_iter = self.fastforward_training(reranker, weights_output_path, loss_fn) if self.config["fastforward"] else 0
        logger.info("starting training from iteration %s/%s", initial_iter, self.config["niters"])

        num_workers = 1 if self.config["multithread"] else 0
        train_dataloader = torch.utils.data.DataLoader(
            train_dataset, batch_size=self.config["batch"], pin_memory=True, num_workers=num_workers
        )
        # dataiter = iter(train_dataloader)
        # sample_input = dataiter.next()
        # summary_writer.add_graph(
        #     reranker.model,
        #     [
        #         sample_input["query"].to(self.device),
        #         sample_input["posdoc"].to(self.device),
        #         sample_input["negdoc"].to(self.device),
        #     ],
        # )

        train_loss = []
        # are we resuming training?
        if initial_iter > 0:
            train_loss = self.load_loss_file(loss_fn)

            # are we done training?
            if initial_iter < self.config["niters"]:
                logger.debug("fastforwarding train_dataloader to iteration %s", initial_iter)
                batches_per_epoch = self.config["itersize"] // self.config["batch"]
                for niter in range(initial_iter):
                    for bi, batch in enumerate(train_dataloader):
                        if (bi + 1) % batches_per_epoch == 0:
                            break

        dev_best_metric = -np.inf
        validation_frequency = self.config["validatefreq"]
        train_start_time = time.time()
        for niter in range(initial_iter, self.config["niters"]):
            model.train()

            iter_start_time = time.time()
            iter_loss_tensor = self.single_train_iteration(reranker, train_dataloader)
            logger.info("A single iteration takes {}".format(time.time() - iter_start_time))
            train_loss.append(iter_loss_tensor.item())
            logger.info("iter = %d loss = %f", niter, train_loss[-1])

            # write model weights to file
            weights_fn = weights_output_path / f"{niter}.p"
            reranker.save_weights(weights_fn, self.optimizer)
            # predict performance on dev set

            if niter % validation_frequency == 0:
                pred_fn = dev_output_path / f"{niter}.run"
                preds = self.predict(reranker, dev_data, pred_fn)

                # log dev metrics
                metrics = evaluator.eval_runs(preds, qrels, evaluator.DEFAULT_METRICS, relevance_level)
                logger.info("dev metrics: %s", " ".join([f"{metric}={v:0.3f}" for metric, v in sorted(metrics.items())]))
                summary_writer.add_scalar("ndcg_cut_20", metrics["ndcg_cut_20"], niter)
                summary_writer.add_scalar("map", metrics["map"], niter)
                summary_writer.add_scalar("P_20", metrics["P_20"], niter)
                # write best dev weights to file
                if metrics[metric] > dev_best_metric:
                    reranker.save_weights(dev_best_weight_fn, self.optimizer)

            # write train_loss to file
            loss_fn.write_text("\n".join(f"{idx} {loss}" for idx, loss in enumerate(train_loss)))

            summary_writer.add_scalar("training_loss", iter_loss_tensor.item(), niter)
            reranker.add_summary(summary_writer, niter)
            summary_writer.flush()
        logger.info("training loss: %s", train_loss)
        logger.info("Training took {}".format(time.time() - train_start_time))
        summary_writer.close()

        # TODO should we write a /done so that training can be skipped if possible when fastforward=False? or in Task?

    def load_best_model(self, reranker, train_output_path):
        pass

    def predict(self, reranker, pred_data, pred_fn):
        """Predict query-document scores on `pred_data` using `model` and write a corresponding run file to `pred_fn`

        Args:
           model (Reranker): a PyTorch Reranker
           pred_data (IterableDataset): data to predict on
           pred_fn (Path): path to write the prediction run file to

        Returns:
           TREC Run

        """

        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        # save to pred_fn
        model = reranker.model.to(self.device)
        model.eval()

        preds = {}
        num_workers = 1 if self.config["multithread"] else 0
        pred_dataloader = torch.utils.data.DataLoader(
            pred_data, batch_size=self.config["batch"], pin_memory=True, num_workers=num_workers
        )
        with torch.autograd.no_grad():
            for batch in tqdm(pred_dataloader, desc="Predicting", total=len(pred_data) // self.config["batch"]):
                if len(batch["qid"]) != self.config["batch"]:
                    batch = self.fill_incomplete_batch(batch)

                batch = {k: v.to(self.device) if not isinstance(v, list) else v for k, v in batch.items()}
                scores = reranker.test(batch)
                scores = scores.view(-1).cpu().numpy()
                for qid, docid, score in zip(batch["qid"], batch["posdocid"], scores):
                    # Need to use float16 because pytrec_eval's c function call crashes with higher precision floats
                    preds.setdefault(qid, {})[docid] = score.astype(np.float16).item()

        os.makedirs(os.path.dirname(pred_fn), exist_ok=True)
        Searcher.write_trec_run(preds, pred_fn)

        return preds

    def fill_incomplete_batch(self, batch):
        """
        If a batch is incomplete (i.e shorter than the desired batch size), this method fills in the batch with some data.
        How the data is chosen:
        If the values are just a simple list, use the first element of the list to pad the batch
        If the values are tensors/numpy arrays, use repeat() along the batch dimension
        """
        # logger.debug("filling in an incomplete batch")
        repeat_times = math.ceil(self.config["batch"] / len(batch["qid"]))
        diff = self.config["batch"] - len(batch["qid"])

        def pad(v):
            if isinstance(v, np.ndarray) or torch.is_tensor(v):
                _v = v.repeat((repeat_times,) + tuple([1 for x in range(len(v.shape) - 1)]))
            else:
                _v = v + [v[0]] * diff

            return _v[: self.config["batch"]]

        batch = {k: pad(v) for k, v in batch.items()}
        return batch



import os
import pickle

from capreolus import Dependency, ModuleBase


class Reranker(ModuleBase):
    """Base class for Reranker modules. The purpose of a Reranker is to predict relevance scores for input documents. Rerankers are generally supervised methods implemented in PyTorch or TensorFlow.

    Modules should provide:
        - a ``build_model`` method that initializes the model used
        - a ``score`` and a ``test`` method that take a representation created by an :class:`~capreolus.extractor.Extractor` module as input and return document scores
        - a ``load_weights`` and a ``save_weights`` method, if the base class' PyTorch methods cannot be used
    """

    module_type = "reranker"
    dependencies = [
        Dependency(key="extractor", module="extractor", name="embedtext"),
        Dependency(key="trainer", module="trainer", name="pytorch"),
    ]

    def add_summary(self, summary_writer, niter):
        """
        Write to the summay_writer custom visualizations/data specific to this reranker
        """
        for name, weight in self.model.named_parameters():
            summary_writer.add_histogram(name, weight.data.cpu(), niter)
            # summary_writer.add_histogram(f'{name}.grad', weight.grad, niter)

    def save_weights(self, weights_fn, optimizer):
        if not os.path.exists(os.path.dirname(weights_fn)):
            os.makedirs(os.path.dirname(weights_fn))

        d = {k: v for k, v in self.model.state_dict().items() if ("embedding.weight" not in k and "_nosave_" not in k)}
        with open(weights_fn, "wb") as outf:
            pickle.dump(d, outf, protocol=-1)

        optimizer_fn = weights_fn.as_posix() + ".optimizer"
        with open(optimizer_fn, "wb") as outf:
            pickle.dump(optimizer.state_dict(), outf, protocol=-1)

    def load_weights(self, weights_fn, optimizer):
        with open(weights_fn, "rb") as f:
            d = pickle.load(f)

        cur_keys = set(k for k in self.model.state_dict().keys() if not ("embedding.weight" in k or "_nosave_" in k))
        missing = cur_keys - set(d.keys())
        if len(missing) > 0:
            raise RuntimeError("loading state_dict with keys that do not match current model: %s" % missing)

        self.model.load_state_dict(d, strict=False)

        optimizer_fn = weights_fn.as_posix() + ".optimizer"
        with open(optimizer_fn, "rb") as f:
            optimizer.load_state_dict(pickle.load(f))


from profane import import_all_modules


import_all_modules(__file__, __package__)
